#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Functions and classes for treating nested sampling exploration as a tree.

The root represents the prior volume, branches and sub-branches split the volume.
The leaves of the tree are the integration tail.

Nested sampling proceeds as a breadth first graph search,
with active nodes sorted by likelihood value.
The number of live points are the number of parallel edges (active nodes to do).

Most functions receive the argument "roots", which are the
children of the tree root (main branches).

The exploration is bootstrap-capable without requiring additional
computational effort: The roots are indexed, and the bootstrap explorer
can ignore the rootids it does not know about.


"""

from __future__ import print_function, division
import numpy as np
from numpy import log, log1p, exp, logaddexp
import math
import sys
from .utils import resample_equal
from .ordertest import UniformOrderAccumulator


class TreeNode(object):
    """Simple tree node."""

    def __init__(self, value=None, id=None, children=None):
        """Define TreeNode.

        Parameters
        ----------
        value:
            used to order nodes
        id: int
            refers to the order of discovery and storage (PointPile)
        children: list of :class:TreeNode objects
            children nodes. if None, a empty list is used.
        """
        self.value = value
        self.id = id
        if children is None:
            self.children = []
        else:
            self.children = children

    def __str__(self, indent=0):
        """Visual representation of the node and its children (recursive)."""
        return ' ' * indent + '- Node: %s\n' % self.value + '\n'.join(
            [c.__str__(indent=indent + 2) for c in self.children])

    def __lt__(self, other):
        """Define order of node based on value attribute."""
        return self.value < other.value


class BreadthFirstIterator(object):
    """Generator exploring the tree.

    Nodes are ordered by value and expanded in order.
    The number of edges passing the node "in parallel" are "active".
    """

    def __init__(self, roots):
        """Start with initial set of nodes *roots*."""
        self.roots = roots
        self.reset()

    def reset(self):
        """(Re)start exploration from the top."""
        self.active_nodes = list(self.roots)
        self.active_root_ids = np.arange(len(self.active_nodes))
        self.active_node_values = np.array([n.value for n in self.active_nodes])
        self.active_node_ids = np.array([n.id for n in self.active_nodes])
        assert len(self.active_nodes) == len(self.active_root_ids)
        assert len(self.active_nodes) == len(self.active_node_values)
        # print("starting live points from %d roots" % len(self.roots), len(self.active_nodes))

    def next_node(self):
        """Get next node in order.

        Does not remove the node from active set.

        Returns
        --------
        tuple or None:
            None if done.
            `rootid, node, (active_nodes, active_root_ids, active_node_values, active_node_ids)`
            otherwise

        """
        if self.active_nodes == []:
            return None
        self.next_index = np.argmin(self.active_node_values)
        i = self.next_index
        node = self.active_nodes[i]
        rootid = self.active_root_ids[i]
        assert not isinstance(rootid, float)
        # print("consuming %.1f" % node.value, len(node.children), 'nlive:', len(self.active_nodes))
        assert len(self.active_nodes) == len(self.active_root_ids)
        assert len(self.active_nodes) == len(self.active_node_values)
        return rootid, node, (self.active_nodes, self.active_root_ids, self.active_node_values, self.active_node_ids)

    def drop_next_node(self):
        """Forget the current node."""
        i = self.next_index
        mask = np.ones(len(self.active_nodes), dtype=bool)
        mask[i] = False
        self.active_nodes.pop(i)
        self.active_node_values = self.active_node_values[mask]
        self.active_root_ids = self.active_root_ids[mask]
        self.active_node_ids = self.active_node_ids[mask]
        assert len(self.active_nodes) == len(self.active_root_ids)
        assert len(self.active_nodes) == len(self.active_node_values)

    def expand_children_of(self, rootid, node):
        """Replace the current node with its children in the iterators list of active nodes.

        Parameters
        ----------
        rootid: int
            index of the root returned by the most recent call to :pyfunc:BreadthFirstIterator.next_node
        node: :pyclass:TreeNode
            node returned by the most recent call to :pyfunc:BreadthFirstIterator.next_node
        """
        # print("replacing %.1f" % node.value, len(node.children))
        i = self.next_index
        newnnodes = len(self.active_nodes) - 1 + len(node.children)
        if len(node.children) == 1:
            self.active_nodes[i] = node.children[0]
            self.active_node_values[i] = node.children[0].value
            self.active_root_ids[i] = rootid
            self.active_node_ids[i] = node.children[0].id
        else:
            mask = np.ones(len(self.active_nodes), dtype=bool)
            mask[i] = False
            self.active_nodes.pop(i)
            if len(node.children) == 0:
                self.active_node_values = self.active_node_values[mask]
                self.active_root_ids = self.active_root_ids[mask]
                self.active_node_ids = self.active_node_ids[mask]
            else:
                self.active_nodes += node.children
                self.active_node_values = np.concatenate((self.active_node_values[mask], [c.value for c in node.children]))
                # print(self.active_root_ids, '+', [rootid for c in node.children], '-->')
                self.active_root_ids = np.concatenate((self.active_root_ids[mask], [rootid for c in node.children]))
                self.active_node_ids = np.concatenate((self.active_node_ids[mask], [c.id for c in node.children]))
                # print(self.active_root_ids)
            assert len(self.active_nodes) == len(self.active_root_ids)
            assert len(self.active_nodes) == len(self.active_node_values)
            assert len(self.active_nodes) == len(self.active_node_ids)
        assert newnnodes == len(self.active_nodes), (len(self.active_nodes), newnnodes, len(node.children))
        assert newnnodes == len(self.active_root_ids), (len(self.active_root_ids), newnnodes, len(node.children))
        assert newnnodes == len(self.active_node_values), (len(self.active_node_values), newnnodes, len(node.children))
        assert newnnodes == len(self.active_node_ids), (len(self.active_node_ids), newnnodes, len(node.children))


def _stringify_lanes(lanes, char='║'):
    """unicode-draw lanes, fill with vertical stripes or spaces."""
    return ''.join([' ' if n is None else char for n in lanes])


def print_tree(roots, title='Tree:'):
    """Print a pretty yet compact graphic of the tree.

    Parameters
    ----------
    roots: list of :pyclass:TreeNode
        tree
    """
    print()
    print(title)
    explorer = BreadthFirstIterator(roots)
    lanes = list(roots)
    lastlane = -1

    while True:
        next_node = explorer.next_node()
        if next_node is None:
            break

        rootid, node, (active_nodes, active_rootids, active_values, active_nodeids) = next_node
        laneid = lanes.index(node)
        nchildren = len(node.children)
        leftstr = _stringify_lanes(lanes[:laneid])
        rightstr = _stringify_lanes(lanes[laneid + 1:])

        if lastlane == laneid:
            sys.stdout.write(leftstr + '║' + rightstr + "\n")
        rightstr = rightstr + " \t" + str(node.value)
        if nchildren == 0:
            sys.stdout.write(leftstr + 'O' + rightstr + "\n")
            lanes[laneid] = None  # keep lane empty
        elif nchildren == 1:
            sys.stdout.write(leftstr + '+' + rightstr + "\n")
            lanes[laneid] = node.children[0]
        else:
            # expand width:
            for j, child in enumerate(node.children):
                rightstr2 = _stringify_lanes(lanes[laneid + 1:], char='\\')
                if len(rightstr2) != 0:
                    sys.stdout.write(leftstr + '║' + ' ' * j + rightstr2 + "\n")
            sys.stdout.write(leftstr + '╠' + '╦' * (nchildren - 2) + '╗' + rightstr + "\n")

            lanes.pop(laneid)
            for j, child in enumerate(node.children):
                lanes.insert(laneid, child)
        explorer.expand_children_of(rootid, node)
        lastlane = laneid


def dump_tree(filename, roots, pointpile):
    """Write a copy of the tree to a HDF5 file.

    Parameters
    ----------
    filename: str
        output filename
    roots: list of :pyclass:TreeNode
        tree to store
    pointpile: :class:PointPile
        information on the node points
    """
    import h5py

    nodes_from_ids = []
    nodes_to_ids = []
    nodes_values = []

    explorer = BreadthFirstIterator(roots)
    while True:
        next_node = explorer.next_node()
        if next_node is None:
            break
        rootid, node, (active_nodes, active_rootids, active_values, active_nodeids) = next_node
        for c in node.children:
            nodes_from_ids.append(node.id)
            nodes_to_ids.append(c.id)
            nodes_values.append(c.value)
        explorer.expand_children_of(rootid, node)

    with h5py.File(filename, 'w') as f:
        f.create_dataset('unit_points', data=pointpile.us[:pointpile.nrows,:], compression='gzip', shuffle=True)
        f.create_dataset('points', data=pointpile.ps[:pointpile.nrows,:], compression='gzip', shuffle=True)

        f.create_dataset('nodes_parent_id', data=nodes_from_ids, compression='gzip', shuffle=True)
        f.create_dataset('nodes_child_id', data=nodes_to_ids, compression='gzip', shuffle=True)
        f.create_dataset('nodes_child_logl', data=nodes_values, compression='gzip', shuffle=True)


def count_tree(roots):
    """Return the total number of nodes and maximum number of parallel edges.

    Parameters
    ----------
    roots: list of :pyclass:TreeNode
        tree

    Returns
    --------
    count: int
        total number of nodes
    maxwidth: int
        maximum number of active/parallel nodes encountered
    """
    explorer = BreadthFirstIterator(roots)
    nnodes = 0
    maxwidth = 0

    while True:
        next_node = explorer.next_node()
        if next_node is None:
            return nnodes, maxwidth
        rootid, node, (active_nodes, active_rootids, active_values, active_nodeids) = next_node
        maxwidth = max(maxwidth, len(active_rootids))
        nnodes += 1
        explorer.expand_children_of(rootid, node)


def count_tree_between(roots, lo, hi):
    """Compute basic statistics about a tree.

    Return the total number of nodes and maximum number of parallel edges,
    but only considering a interval of the tree.

    Parameters
    ----------
    roots: list of :pyclass:TreeNode
        tree
    lo: float
        lower value threshold
    hi: float
        upper value threshold

    Returns
    --------
    nnodes: int
        total number of nodes in the value interval lo .. hi (inclusive).
    maxwidth: int
        maximum number of parallel edges

    """
    explorer = BreadthFirstIterator(roots)
    nnodes = 0
    maxwidth = 0

    while True:
        next_node = explorer.next_node()
        if next_node is None:
            return nnodes, maxwidth

        rootid, node, (active_nodes, active_rootids, active_values, active_nodeids) = next_node

        if node.value > hi:
            # can stop already
            return nnodes, maxwidth

        if lo <= node.value <= hi:
            maxwidth = max(maxwidth, len(active_rootids))
            nnodes += 1

        explorer.expand_children_of(rootid, node)


def find_nodes_before(root, value):
    """Identify all nodes that have children above *value*.

    If a root child is above the value, its parent (*root*) is the leaf returned.

    Parameters
    ----------
    root: :pyclass:TreeNode
        tree
    value: float
        selection threshold

    Returns
    --------
    list_of_parents: list of nodes
        parents
    list_of_nforks: list of floats
        The list of number of forks experienced is:
        1 if direct descendent of one of the root node's children,
        where no node had more than one child.
        12 if the root child had 4 children, one of which had 3 children.

    """
    roots = root.children
    parents = []
    parent_weights = []

    weights = {n.id: 1. for n in roots}
    explorer = BreadthFirstIterator(roots)
    while True:
        next_node = explorer.next_node()
        if next_node is None:
            break
        rootid, node, _ = next_node
        if node.value >= value:
            # already past (root child)
            parents.append(root)
            parent_weights.append(1)
            break
        elif any(n.value >= value for n in node.children):
            # found matching parent
            parents.append(node)
            parent_weights.append(weights[node.id])
            explorer.drop_next_node()
        else:
            # continue exploring
            explorer.expand_children_of(rootid, node)
            weights.update({n.id: weights[node.id] * len(node.children)
                            for n in node.children})
        del weights[node.id]
    return parents, parent_weights


class PointPile(object):
    """A in-memory linearized storage of point coordinates.

    :pyclass:TreeNodes only store the logL value and id,
    which is the index in the point pile. The point pile stores
    the point coordinates.
    """

    def __init__(self, udim, pdim, chunksize=1000):
        """Set up point pile.

        Parameters
        -----------
        udim: int
            number of parameters, dimension of unit cube points
        pdim: int
            number of physical (and derived) parameters
        chunksize: int
            the point pile grows as needed, in these intervals.

        """
        self.nrows = 0
        self.chunksize = chunksize
        self.us = np.zeros((self.chunksize, udim))
        self.ps = np.zeros((self.chunksize, pdim))
        self.udim = udim
        self.pdim = pdim

    def add(self, newpointu, newpointp):
        """Save point.

        Parameters
        -----------
        newpointu: array
            point (in u-space)
        newpointp: array
            point (in p-space)

        Returns
        ---------
        index: int
            index of the new point in the pile
        """
        if self.nrows >= self.us.shape[0]:
            self.us = np.concatenate((self.us, np.zeros((self.chunksize, self.udim))))
            self.ps = np.concatenate((self.ps, np.zeros((self.chunksize, self.pdim))))
        assert len(newpointu) == self.us.shape[1], (newpointu, self.us.shape)
        assert len(newpointp) == self.ps.shape[1], (newpointp, self.ps.shape)
        self.us[self.nrows,:] = newpointu
        self.ps[self.nrows,:] = newpointp
        self.nrows += 1
        return self.nrows - 1

    def getu(self, i):
        """Get cube point(s) with index(indices) `i`."""
        return self.us[i]

    def getp(self, i):
        """Get parameter point(s) with index(indices) `i`."""
        return self.ps[i]

    def make_node(self, value, u, p):
        """Store point in pile, and create a new tree node that points to it.

        Parameters
        -----------
        value: float
            value to store in node (loglikelihood)
        u: array
            point (in u-space)
        p: array
            point (in p-space)

        Returns
        ---------
        node: :pyclass:TreeNode
            node
        """
        index = self.add(u, p)
        return TreeNode(value=value, id=index)


class SingleCounter(object):
    """Evidence log(Z) and posterior weight summation for a Nested Sampling tree."""

    def __init__(self, random=False):
        """Initialise counter.

        Parameters
        ----------
        random: bool
            if False, use mean estimator for volume shrinkage
            if True, draw a random sample

        """
        self.reset()
        self.random = random

    def reset(self):
        """Reset counters and integration."""
        self.logweights = []
        self.H = None
        self.logZ = -np.inf
        self.logZerr = np.inf
        self.logVolremaining = 0
        self.i = 0
        self.fraction_remaining = np.inf
        self.Lmax = -np.inf

    @property
    def logZremain(self):
        """Estimate conservatively the logZ of the current tail (un-opened nodes)."""
        return self.Lmax + self.logVolremaining

    def passing_node(self, node, parallel_nodes):
        """Accumulate node to the integration.

        Parameters
        -----------
        node: TreeNode
            breadth-first removed node
        parallel_nodes: list
            nodes active next to node

        """
        # node is being consumed
        # we have parallel arcs to parallel_nodes

        nchildren = len(node.children)
        Li = node.value
        nlive = len(parallel_nodes)

        if nchildren >= 1:
            # one arc terminates, another is spawned

            # weight is the size of the slice off the volume
            logleft = log1p(-exp(-1. / nlive))
            logright = -1. / nlive
            if self.random:
                randompoint = np.random.beta(1, nlive)
                logleft = log(randompoint)
                logright = log1p(-randompoint)

            logwidth = logleft + self.logVolremaining
            wi = logwidth + Li
            self.logweights.append(logwidth)
            if math.isinf(self.logZ):
                self.logZ = wi
                self.H = Li - self.logZ
            else:
                logZnew = logaddexp(self.logZ, wi)
                self.H = exp(wi - logZnew) * Li + exp(self.logZ - logZnew) * (self.H + self.logZ) - logZnew
                assert np.all(np.isfinite(self.H)), (self.H, wi, logZnew, Li, self.logZ)
                self.logZ = logZnew

            # print(self.H)
            # self.Lmax = max(node.value, self.Lmax)
            # self.Lmax = max((n.value for n in parallel_nodes))
            # logZremain = parallel_nodes.max() + self.logVolremaining
            # print("L=%.1f N=%d V=%.2e logw=%.2e logZ=%.1f logZremain=%.1f" % (Li, nlive, self.logVolremaining, wi, self.logZ, logZremain))
            # volume is reduced by exp(-1/N)
            self.logVolremaining += logright
            # TODO: this needs to change if nlive varies
            if self.H >= 0:
                self.logZerr = (self.H / nlive)**0.5
                assert np.isfinite(self.logZerr), (self.H, nlive)
        else:
            # contracting!

            # weight is simply volume / Nlive
            logwidth = self.logVolremaining - log(nlive)
            wi = logwidth + Li

            self.logweights.append(logwidth)
            self.logZ = logaddexp(self.logZ, wi)

            # print("L=%.1f N=%d V=%.2e logw=%.2e logZ=%.1f" % (Li, nlive, self.logVolremaining, wi, self.logZ))

            # the volume shrinks by (N - 1) / N
            # self.logVolremaining += log(1 - exp(-1. / nlive))
            # if nlive = 1, we are removing the last point, so remaining
            # volume is zero (leads to log of -inf, as expected)
            with np.errstate(divide='ignore'):
                self.logVolremaining += log1p(-1.0 / nlive)


class MultiCounter(object):
    """Like SingleCounter, but bootstrap capable.

    **Attributes**:

    - ``logZ``, ``logZerr``, ``logVolremaining``: main estimator
      ``logZerr`` is probably not reliable, because it needs ``nlive``
      to convert ``H`` to ``logZerr``.
    - ``Lmax``: highest loglikelihood currently known
    - ``logZ_bs``, ``logZerr_bs``: bootstrapped logZ estimate
    - ``logZremain``, ``remainder_ratio``: weight and fraction of the unexplored remainder

    Each of the following has as many entries as number of iterations:

    - ``all_H``, ``all_logZ``, ``all_logVolremaining``, ``logweights``:
      information for all instances
      first entry is the main estimator, i.e., not bootstrapped
    - ``istail``: whether that node was a leaf.
    - ``nlive``: number of parallel arcs ("live points")

    """

    def __init__(self, nroots, nbootstraps=10, random=False, check_insertion_order=False):
        """Initialise counter.

        Parameters
        ----------
        nroots: int
            number of children the tree root has
        nbootstraps: int
            number of bootstrap rounds
        random: bool
            if False, use mean estimator for volume shrinkage
            if True, draw a random sample

        """
        allyes = np.ones(nroots, dtype=bool)
        # the following is a masked array of size (nbootstraps+1, nroots)
        # which rootids are active in each bootstrap instance
        # the first one contains everything
        self.rootids = [allyes]
        self.insertion_order_sample = []
        # np.random.seed(1)
        for i in range(nbootstraps):
            mask = ~allyes
            rootids = np.unique(np.random.randint(nroots, size=nroots))
            mask[rootids] = True
            self.rootids.append(mask)
        self.rootids = np.array(self.rootids)
        self.random = random
        self.ncounters = len(self.rootids)

        self.check_insertion_order = check_insertion_order
        self.insertion_order_threshold = 4
        self.insertion_order_accumulator = UniformOrderAccumulator()

        self.reset(len(self.rootids))

    def reset(self, nentries):
        """Reset counters/integrator."""
        self.logweights = []
        self.istail = []
        self.logZ = -np.inf
        self.logZerr = np.inf
        self.all_H = -np.nan * np.ones(nentries)
        self.all_logZ = -np.inf * np.ones(nentries)
        self.all_logVolremaining = np.zeros(nentries)
        self.logVolremaining = 0.0
        self.Lmax = -np.inf

        self.all_logZremain = np.inf * np.ones(nentries)
        self.logZremainMax = np.inf
        self.logZremain = np.inf
        self.remainder_ratio = 1.0
        self.remainder_fraction = 1.0

        self.insertion_order_accumulator.reset()
        self.insertion_order_runs = []

    @property
    def logZ_bs(self):
        """Estimate logZ from the bootstrap ensemble."""
        return self.all_logZ[1:].mean()

    @property
    def logZerr_bs(self):
        """Estimate logZ error from the bootstrap ensemble."""
        return self.all_logZ[1:].std()

    @property
    def insertion_order_runlength(self):
        """Get shortest insertion order test run.

        Returns
        --------
        shortest_run_length: int
            Shortest insertion order test run length.

        The MWW (U-test) statistic is considered at each iteration.
        When it exceeds a threshold (4 sigma by default, `insertion_order_threshold`),
        the statistic is reset. The run length is recorded.
        This property returns the shortest run length of all recorded
        so far, or infinity otherwise.

        At 4 sigma, run lengths no shorter than 10^5.5 are expected
        in unbiased runs.
        """
        runs = self.insertion_order_runs
        if len(runs) == 0:
            return np.inf
        else:
            return min(runs)

    @property
    def insertion_order_converged(self):
        """Check convergence.

        Returns
        --------
        converged: bool
            Whether the run is unbiased according to a U-test.

        The MWW (U-test) statistic is considered at each iteration.
        When it exceeds a threshold (4 sigma by default, `insertion_order_threshold`),
        the statistic is reset. The run length is recorded.
        This property returns the shortest run length of all recorded
        so far, or infinity otherwise.

        At 4 sigma, run lengths no shorter than 10^5.5 are expected
        in unbiased runs. If the number of runs exceeds the number
        of iterations divided by 10^5.5, the run is likely biased
        and not converged.

        If not converged, the step sampler may need to use more steps,
        or the problem needs to be  reparametrized.
        """
        # we expect run lengths not shorter than 300000 for 4sigma
        # if we get many more than expected from the number of iterations
        # the run has not converged
        niter = len(self.logweights)
        expected_number = max(1, int(np.ceil(niter / 10**(5.5))))
        return len(self.insertion_order_runs) <= expected_number

    def passing_node(self, rootid, node, rootids, parallel_values):
        """Accumulate node to the integration.

        Breadth-first removed `node` and nodes active next to node (`parallel_nodes`).
        rootid and rootids are needed to identify which bootstrap instance
        should accumulate.

        Parameters
        ----------
        rootid: :pyclass:TreeNode
            root node this `node` is from.
        node: :pyclass:TreeNode
            node being processed.
        rootids: array of ints
            for each parallel node, which root it belongs to.
        parallel_values: float array
            loglikelihood values of nodes passing `node`.

        """
        # node is being consumed
        # we have parallel arcs to parallel_nodes

        assert not isinstance(rootid, float)
        nchildren = len(node.children)
        Li = node.value
        # in which bootstraps is rootid?
        active = self.rootids[:,rootid]
        # how many live points does each bootstrap have?
        nlive = self.rootids[:,rootids].sum(axis=1)
        nlive0 = nlive[0]

        if nchildren >= 1:
            # one arc terminates, another is spawned

            # weight is the size of the slice off the volume
            if self.random:
                randompoint = np.random.beta(1, nlive, size=self.ncounters)
                logleft = log(randompoint)
                logright = log1p(-randompoint)
                logleft[0] = log1p(-exp(-1. / nlive0))
                logright[0] = -1. / nlive0
            else:
                logleft = log1p(-exp(-1. / nlive))
                logright = -1. / nlive

            logwidth = logleft + self.all_logVolremaining
            logwidth[~active] = -np.inf
            wi = logwidth[active] + Li
            self.logweights.append(logwidth)
            self.istail.append(False)

            # print("updating continuation...", Li)
            assert active[0], (active, rootid)
            logZ = self.all_logZ[active]
            logZnew = logaddexp(logZ, wi)
            H = exp(wi - logZnew) * Li + exp(logZ - logZnew) * (self.all_H[active] + logZ) - logZnew
            first_setting = np.isnan(H)
            # print()
            # print("Hnext:", H[0], first_setting[0])
            assert np.isfinite(H[~first_setting]).all(), (first_setting, self.all_H[active][~first_setting], H, wi, logZnew, Li, logZ)
            self.all_logZ[active] = np.where(first_setting, wi, logZnew)
            # print("logZ:", self.all_logZ[0])
            if first_setting[0]:
                assert np.all(np.isfinite(Li - wi)), (Li, wi)
            else:
                assert np.isfinite(self.all_H[0]), self.all_H[0]
                assert np.isfinite(H[0]), (first_setting[0], H[0], self.all_H[0], wi[0], logZnew[0], Li, logZ[0])
            self.all_H[active] = np.where(first_setting, -logwidth[active], H)
            # print("H:", self.all_H)
            assert np.isfinite(self.all_H[active]).all(), (self.all_H[active], first_setting[0], H[0], self.all_H[0], wi[0], logZnew[0], Li, logZ[0])
            # assert np.all(np.isfinite(self.all_H[active])), (H, self.all_H[active], wi, logZnew, Li, logZ)
            self.logZ = self.all_logZ[0]
            assert np.all(np.isfinite(self.all_logZ[active])), (self.all_logZ[active])

            # self.Lmax = max((n.value for n in parallel_nodes))
            # print("L=%.1f N=%d V=%.2e logw=%.2e logZ=%.1f logZremain=%.1f" % (
            #  Li, nlive[0], self.logVolremaining, wi[0], self.logZ, logZremain))
            # print("L=%.1f N=%d V=%.2e logw=%.2e logZ=%.1f logZremain=%.1f" % (
            #  Li, nlive[0], self.all_logVolremaining[0], (logwidth + Li)[0], self.all_logZ[0], logZremain))
            # print("L=%.1f N=%d V=%.2e logw=%.2e logZ=<%.1f logZremain=%.1f" % (
            #  Li, nlive[1], self.all_logVolremaining[1], (logwidth + Li)[1], self.all_logZ[1], logZremain))

            if self.all_H[0] > 0:
                # TODO: this needs to change if nlive varies
                self.logZerr = (self.all_H[0] / nlive0)**0.5
            # assert np.all(np.isfinite(self.logZerr)), (self.logZerr, self.all_H[0], nlive)

            # volume is reduced by exp(-1/N)
            self.all_logVolremaining[active] += logright[active]
            self.logVolremaining = self.all_logVolremaining[0]

            if self.check_insertion_order and len(np.unique(parallel_values)) == len(parallel_values):
                acc = self.insertion_order_accumulator
                parallel_values_here = parallel_values[self.rootids[0, rootids]]
                for child in node.children:
                    # rootids is 400 ints pointing to the root id where each parallel_values is from
                    # self.rootids[i] says which rootids belong to this bootstrap
                    # need which of the parallel_values are active here
                    acc.add((parallel_values_here < child.value).sum(), nlive0)
                    if abs(acc.zscore) > self.insertion_order_threshold:
                        self.insertion_order_runs.append(len(acc))
                        acc.reset()

        else:
            # contracting!
            # print("contracting...", Li)

            # weight is simply volume / Nlive
            logwidth = -np.inf * np.ones(self.ncounters)
            logwidth[active] = self.all_logVolremaining[active] - log(nlive[active])
            wi = logwidth + Li

            self.logweights.append(logwidth)
            self.istail.append(True)
            self.all_logZ[active] = logaddexp(self.all_logZ[active], wi[active])
            self.logZ = self.all_logZ[0]

            # print("L=%.1f N=%d V=%.2e logw=%.2e logZ=%.1f" % (Li, nlive, self.logVolremaining, wi, self.logZ))

            # the volume shrinks by (N - 1) / N
            # self.logVolremaining += log(1 - exp(-1. / nlive))
            # if nlive = 1, we are removing the last point, so remaining
            # volume is zero (leads to log of -inf, as expected)
            with np.errstate(divide='ignore'):
                self.all_logVolremaining[active] += log1p(-1.0 / nlive[active])
            self.logVolremaining = self.all_logVolremaining[0]

        V = self.all_logVolremaining - log(nlive0)
        Lmax = np.max(parallel_values)
        self.all_logZremain = V + log(np.sum(exp(parallel_values - Lmax))) + Lmax
        self.logZremainMax = self.all_logZremain.max()
        self.logZremain = self.all_logZremain[0]
        with np.errstate(over='ignore', under='ignore'):
            self.remainder_ratio = exp(self.logZremain - self.logZ)
            self.remainder_fraction = 1.0 / (1 + exp(self.logZ - self.logZremain))


def combine_results(saved_logl, saved_nodeids, pointpile, main_iterator, mpi_comm=None):
    """Combine a sequence of likelihoods and nodes into a summary dictionary.

    Parameters
    ----------
    saved_logl: list of floats
        loglikelihoods of dead points
    saved_nodeids: list of ints
        indices of dead points
    pointpile: :pyclass:PointPile
        Point pile.
    main_iterator: :pyclass:BreadthFirstIterator
        iterator used
    mpi_comm:
        MPI communicator object, or None if MPI is not used.

    Returns
    --------
    results: dict
        All information of the run. Important keys:
        Number of nested sampling iterations (niter),
        Evidence estimate (logz),
        Effective Sample Size (ess),
        H (information gain),
        weighted samples (weighted_samples),
        equally weighted samples (samples),
        best-fit point information (maximum_likelihood),
        posterior summaries (posterior).
        The rank order test score (insertion_order_MWW_test) is
        included if the iterator has it.
    """
    assert np.shape(main_iterator.logweights) == (len(saved_logl), len(main_iterator.all_logZ)), (
        np.shape(main_iterator.logweights),
        np.shape(saved_logl),
        np.shape(main_iterator.all_logZ))

    saved_logl = np.array(saved_logl)
    saved_u = pointpile.getu(saved_nodeids)
    saved_v = pointpile.getp(saved_nodeids)
    saved_logwt = np.array(main_iterator.logweights)
    saved_logwt0 = saved_logwt[:,0]
    saved_logwt_bs = saved_logwt[:,1:]
    logZ_bs = main_iterator.all_logZ[1:]
    assert len(saved_logwt_bs) == len(saved_nodeids), (saved_logwt_bs.shape, len(saved_nodeids))

    if mpi_comm is not None:
        # spread logZ_bs, saved_logwt_bs
        recv_saved_logwt_bs = mpi_comm.gather(saved_logwt_bs, root=0)
        recv_saved_logwt_bs = mpi_comm.bcast(recv_saved_logwt_bs, root=0)
        saved_logwt_bs = np.concatenate(recv_saved_logwt_bs, axis=1)

        recv_logZ_bs = mpi_comm.gather(logZ_bs, root=0)
        recv_logZ_bs = mpi_comm.bcast(recv_logZ_bs, root=0)
        logZ_bs = np.concatenate(recv_logZ_bs)

    saved_wt_bs = exp(saved_logwt_bs + saved_logl.reshape((-1, 1)) - logZ_bs)
    saved_wt0 = exp(saved_logwt0 + saved_logl - main_iterator.all_logZ[0])

    # compute fraction in tail
    w = saved_wt0 / saved_wt0.sum()
    assert np.isclose(w.sum() - 1, 0), w.sum()
    ess = len(w) / (1.0 + ((len(w) * w - 1)**2).sum() / len(w))
    tail_fraction = w[np.asarray(main_iterator.istail)].sum()
    if tail_fraction != 0:
        logzerr_tail = logaddexp(log(tail_fraction) + main_iterator.logZ, main_iterator.logZ) - main_iterator.logZ
    else:
        logzerr_tail = 0

    logzerr_bs = (logZ_bs - main_iterator.logZ).max()
    logzerr_total = (logzerr_tail**2 + logzerr_bs**2)**0.5
    samples = resample_equal(saved_v, w)

    ndim = saved_u.shape[1]
    information_gain_bits = []
    for i in range(ndim):
        H, _ = np.histogram(saved_u[:,i], weights=saved_wt0, density=True, bins=np.linspace(0, 1, 40))
        information_gain_bits.append(float((np.log2(1 / ((H + 0.001) * 40)) / 40).sum()))

    j = saved_logl.argmax()

    results = dict(
        niter=len(saved_logl),
        logz=main_iterator.logZ, logzerr=logzerr_total,
        logz_bs=logZ_bs.mean(),
        logz_single=main_iterator.logZ,
        logzerr_tail=logzerr_tail,
        logzerr_bs=logzerr_bs,
        ess=ess,
        H=main_iterator.all_H[0], Herr=main_iterator.all_H.std(),
        posterior=dict(
            mean=samples.mean(axis=0).tolist(),
            stdev=samples.std(axis=0).tolist(),
            median=np.percentile(samples, 50, axis=0).tolist(),
            errlo=np.percentile(samples, 15.8655, axis=0).tolist(),
            errup=np.percentile(samples, 84.1345, axis=0).tolist(),
            information_gain_bits=information_gain_bits,
        ),
        weighted_samples=dict(
            upoints=saved_u, points=saved_v, weights=saved_wt0, logw=saved_logwt0,
            bootstrapped_weights=saved_wt_bs, logl=saved_logl),
        samples=samples,
        maximum_likelihood=dict(
            logl=saved_logl[j],
            point=saved_v[j,:].tolist(),
            point_untransformed=saved_u[j,:].tolist(),
        ),
    )

    if getattr(main_iterator, 'check_insertion_order', False):
        results['insertion_order_MWW_test'] = dict(
            independent_iterations=main_iterator.insertion_order_runlength,
            converged=main_iterator.insertion_order_converged,
        )

    return results


def logz_sequence(root, pointpile, nbootstraps=12, random=True, onNode=None, verbose=False, check_insertion_order=True):
    """Run MultiCounter through tree `root`.

    Keeps track of, and returns ``(logz, logzerr, logv, nlive)``.

    Parameters
    ----------
    root: :pyclass:TreeNode
        Tree
    pointpile: :pyclass:PointPile
        Point pile
    nbootstraps: int
        Number of independent iterators
    random: bool
        Whether to randomly draw volume estimates
    onNode: function
        Function to call for every node.
        receives current node and the iterator
    verbose: bool
        Whether to show a progress indicator on stderr
    check_insertion_order: bool
        Whether to perform a rolling insertion order rank test

    Returns
    --------
    results: dict
        Run information, see :pyfunc:combine_results
    sequence: dict
        Each entry of the dictionary is results['niter'] long,
        and contains the state of information at that iteration.
        Important keys are:
        Iteration number (niter),
        Volume estimate (logvol), loglikelihood (logl), absolute logarithmic weight (logwt),
        Relative weight (weights), point (samples),
        Number of live points (nlive),
        Evidence estimate (logz) and its uncertainty (logzerr),
        Rank test score (insert_order).

    """
    roots = root.children

    Lmax = -np.inf

    explorer = BreadthFirstIterator(roots)
    # Integrating thing
    main_iterator = MultiCounter(
        nroots=len(roots), nbootstraps=max(1, nbootstraps),
        random=random, check_insertion_order=check_insertion_order)
    main_iterator.Lmax = max(Lmax, max(n.value for n in roots))

    logz = []
    logzerr = []
    nlive = []
    logvol = []
    niter = 0

    saved_nodeids = []
    saved_logl = []
    insert_order = []
    # we go through each live point (regardless of root) by likelihood value
    while True:
        next_node = explorer.next_node()
        if next_node is None:
            break
        rootid, node, (active_nodes, active_rootids, active_values, active_node_ids) = next_node
        # this is the likelihood level we have to improve upon
        Lmin = node.value

        if onNode:
            onNode(node, main_iterator)

        logz.append(main_iterator.logZ)
        with np.errstate(invalid='ignore'):
            # first time they are all the same
            logzerr.append(main_iterator.logZerr_bs)

        if len(np.unique(active_values)) == len(active_values) and len(node.children) > 0:
            child_insertion_order = (active_values > node.children[0].value).sum()
            insert_order.append(2 * (child_insertion_order + 1.) / len(active_values))
        else:
            insert_order.append(np.nan)

        nlive.append(len(active_values))
        logvol.append(main_iterator.logVolremaining)

        niter += 1
        if verbose:
            sys.stderr.write("%d...\r" % niter)

        saved_logl.append(Lmin)
        saved_nodeids.append(node.id)
        # inform iterators (if it is their business) about the arc
        main_iterator.passing_node(rootid, node, active_rootids, active_values)
        explorer.expand_children_of(rootid, node)

    logwt = np.asarray(saved_logl) + np.asarray(main_iterator.logweights)[:,0]
    logvol[-1] = logvol[-2]

    results = combine_results(saved_logl, saved_nodeids, pointpile, main_iterator)
    sequence = dict(
        logz=np.asarray(logz),
        logzerr=np.asarray(logzerr),
        logvol=np.asarray(logvol),
        samples_n=np.asarray(nlive),
        nlive=np.asarray(nlive),
        insert_order=np.asarray(insert_order),
        logwt=logwt,
        niter=niter,
        logl=saved_logl,
        weights=results['weighted_samples']['weights'],
        samples=results['weighted_samples']['points'],
    )

    return sequence, results



#!/usr/bin/env python
# coding: utf-8

import numpy as np
from ultranest.utils import submasks
from ultranest.stepfuncs import evolve, step_back
from ultranest.stepfuncs import generate_cube_oriented_direction, \
   generate_random_direction, generate_region_oriented_direction, generate_region_random_direction


class PopulationSliceSampler():
    def __init__(
        self, popsize, nsteps, generate_direction, scale=1.0, 
        scale_adapt_factor=0.9, log=False, logfile=None
    ):
        """
        Vectorized slice/HARM sampler.

        Revert until all previous steps have likelihoods allL above Lmin.
        Updates currentt, generation and allL, in-place.

        Parameters
        ----------
        popsize: int
            number of walkers to maintain
        nsteps: int
            number of steps to take until the found point is accepted as independent.
        generate_direction: function `(u, region, scale) -> v`
            function such as `generate_unit_directions`, which
            generates a random slice direction.
        scale: float
            initial guess scale for the length of the slice
        scale_adapt_factor: float
            smoothing factor for updating scale.
            if near 1, scale is barely updating, if near 0,
            the last slice length is used as a initial guess for the next.

        """
        self.nsteps = nsteps
        self.nrejects = 0
        self.scale = scale
        self.scale_adapt_factor = scale_adapt_factor
        self.allu = []
        self.allL = []
        self.currentt = []
        self.currentv = []
        self.currentp = []
        self.generation = []
        self.current_left = []
        self.current_right = []
        self.searching_left = []
        self.searching_right = []
        self.ringindex = 0

        self.log = log
        self.logfile = logfile

        self.popsize = popsize
        self.generate_direction = generate_direction

    def __str__(self):
        return 'PopulationSliceSampler(popsize=%d, nsteps=%d, generate_direction=%s, scale=%.g)' % (
                self.popsize, self.nsteps, self.generate_direction, self.scale)

    def region_changed(self, Ls, region):
        """notification that the region changed. Currently not used."""
        # self.scale = region.us.std(axis=1).mean()
        if self.logfile:
            self.logfile.write("region-update\t%g\t%g\n" % (self.scale, region.us.std(axis=1).mean()))

    def _setup(self, ndim):
        """Allocate arrays."""
        self.allu = np.zeros((self.popsize, self.nsteps + 1, ndim)) + np.nan
        self.allL = np.zeros((self.popsize, self.nsteps + 1)) + np.nan
        self.currentt = np.zeros(self.popsize) + np.nan
        self.currentv = np.zeros((self.popsize, ndim)) + np.nan
        self.generation = np.zeros(self.popsize, dtype=int) - 1
        self.current_left = np.zeros(self.popsize)
        self.current_right = np.zeros(self.popsize)
        self.searching_left = np.zeros(self.popsize, dtype=bool)
        self.searching_right = np.zeros(self.popsize, dtype=bool)

    def step_back(self, Lmin):
        """see `:func:ultranest.stepfuncs.step_back` :func:ultranest.stepfuncs.step_back."""
        step_back(Lmin, self.allL, self.generation, self.currentt)

    def setup_start(self, us, Ls, starting):
        """Initialize walker starting points.

        For iteration zero, randomly selects a live point as starting point.

        Parameters
        ----------
        us: np.array((nlive, ndim))
            live points
        Ls: np.array(nlive)
            loglikelihoods live points
        starting: np.array(nwalkers, dtype=bool)
            which walkers to initialize.

        """
        if self.log: print("setting up:", starting)
        nlive = len(us)
        i = np.random.randint(nlive, size=starting.sum())

        if not starting.all():
            while starting[self.ringindex]:
                # if the one we are waiting for is being restarted,
                # we may as well pick the next one to wait for
                # because every other one is started from a random point
                # as well
                self.shift()

        self.allu[starting,0] = us[i]
        self.allL[starting,0] = Ls[i]
        self.generation[starting] = 0

    @property
    def status(self):
        s1 = ('G:' + ''.join(['%d' % g if g >= 0 else '_' for g in self.generation]))
        s2 = ('S:' + ''.join(['S' if not np.isfinite(self.currentt[i]) else 'L' if self.searching_left[i] else 'R' if self.searching_right[i] else 'B'
            for i in range(self.popsize)]))
        return s1 + '  ' + s2

    def setup_brackets(self, mask_starting, region):
        """Pick starting direction and range for slice

        Parameters
        ----------
        region: MLFriends object
            Region
        mask_starting: np.array(nwalkers, dtype=bool)
            which walkers to set up.

        """
        if self.log: print("starting brackets:", mask_starting)
        i_starting, = np.where(mask_starting)
        self.current_left[i_starting] = -self.scale
        self.current_right[i_starting] = self.scale
        self.searching_left[i_starting] = True
        self.searching_right[i_starting] = True
        self.currentt[i_starting] = 0
        # choose direction for new slice
        self.currentv[i_starting,:] = self.generate_direction(
            self.allu[i_starting, self.generation[i_starting]],
            region)

    def _setup_currentp(self, nparams):
        if self.log: print("setting currentp")
        self.currentp = np.zeros((self.popsize, nparams)) + np.nan

    def advance(self, transform, loglike, Lmin):
        """Advance the walker population

        Parameters
        ----------
        transform: function
            prior transform function
        loglike: function
            loglikelihood function
        Lmin: float
            current log-likelihood threshold

        """
        movable = self.generation < self.nsteps
        all_movable = movable.all()
        # print("moving ", movable.sum(), self.popsize)
        if all_movable:
            i = np.arange(self.popsize)
            args = [
                self.allu[i, self.generation],
                self.allL[i, self.generation],
                # pass values directly
                self.currentt,
                self.currentv,
                self.current_left,
                self.current_right,
                self.searching_left,
                self.searching_right
            ]
            del i
        else:
            args = [
                self.allu[movable, self.generation[movable]],
                self.allL[movable, self.generation[movable]],
                # this makes copies
                self.currentt[movable],
                self.currentv[movable],
                self.current_left[movable],
                self.current_right[movable],
                self.searching_left[movable],
                self.searching_right[movable]
            ]
        if self.log: print("evolve will advance:", movable)

        (
            (
            currentt, currentv,
            current_left, current_right, searching_left, searching_right),
            (success, unew, pnew, Lnew),
            nc
        ) = evolve(transform, loglike, Lmin, *args, log=self.log)

        if self.log: print("movable", movable.shape, movable.sum(), success.shape)
        moved = submasks(movable, success)
        if self.log: print("evolve moved:", moved)
        self.generation[moved] += 1
        if len(pnew) > 0:
            if len(self.currentp) == 0:
                self._setup_currentp(nparams=pnew.shape[1])

            if self.log: print("currentp", self.currentp[moved,:].shape, pnew.shape)
            self.currentp[moved,:] = pnew

        # update with what we learned
        # print(currentu.shape, currentL.shape, success.shape, self.generation[movable])
        self.allu[moved, self.generation[moved]] = unew
        self.allL[moved, self.generation[moved]] = Lnew
        if all_movable:
            # in this case, the values were directly overwritten
            pass
        else:
            self.currentt[movable] = currentt
            self.currentv[movable] = currentv
            self.current_left[movable] = current_left
            self.current_right[movable] = current_right
            self.searching_left[movable] = searching_left
            self.searching_right[movable] = searching_right
        return nc

    def shift(self):
        """Update walker from which to pick next."""
        # this is a ring buffer
        # shift index forward, wrapping around
        # this is better than copying memory around when a element is removed
        self.ringindex = (self.ringindex + 1) % self.popsize

    def __next__(
        self, region, Lmin, us, Ls, transform, loglike, ndraw=10,
        plot=False, tregion=None, log=False
    ):
        """Sample a new live point.

        Parameters
        ----------
        region: MLFriends object
            Region
        Lmin: float
            current log-likelihood threshold
        us: np.array((nlive, ndim))
            live points
        Ls: np.array(nlive)
            loglikelihoods live points
        transform: function
            prior transform function
        loglike: function
            loglikelihood function
        ndraw: int
            not used
        plot: bool
            not used
        tregion: bool
            not used
        log: bool
            not used

        Returns
        -------
        u: np.array(ndim) or None
            new point coordinates (None if not yet available)
        p: np.array(nparams) or None
            new point transformed coordinates (None if not yet available)
        L: float or None
            new point likelihood (None if not yet available)
        nc: int

        """
        nlive, ndim = us.shape
        # initialize
        if len(self.allu) == 0:
            self._setup(ndim)

        #print(str(self), "(start)")
        self.step_back(Lmin)

        starting = self.generation < 0
        if starting.any():
            self.setup_start(us[Ls > Lmin], Ls[Ls > Lmin], starting)
        assert (self.generation >= 0).all(), self.generation

        #if self.log: print("generation:", self.generation)

        # find those where bracket is undefined:
        mask_starting = ~np.isfinite(self.currentt)
        if mask_starting.any():
            self.setup_brackets(mask_starting, region)

        if self.log: print(str(self), "(before)")
        nc = self.advance(transform, loglike, Lmin)
        if self.log: print(str(self), "(after)")

        # harvest top individual if possible
        if self.generation[self.ringindex] == self.nsteps:
            if self.log: print("have a candidate")
            u, p, L = self.allu[self.ringindex, self.nsteps, :].copy(), self.currentp[self.ringindex, :].copy(), self.allL[self.ringindex, self.nsteps].copy()
            assert np.isfinite(u).all(), u
            assert np.isfinite(p).all(), p
            self.generation[self.ringindex] = -1
            self.currentt[self.ringindex] = np.nan
            self.allu[self.ringindex,:,:] = np.nan
            self.allL[self.ringindex,:] = np.nan

            # adjust guess length
            newscale = (self.current_right[self.ringindex] - self.current_left[self.ringindex]) / 2
            self.scale = self.scale * 0.9 + 0.1 * newscale

            self.shift()
            return u, p, L, nc
        else:
            return None, None, None, nc



from __future__ import print_function, division
import os
import numpy as np
from ultranest.store import TextPointStore
from ultranest.netiter import PointPile, TreeNode, count_tree, print_tree, dump_tree
from ultranest.netiter import SingleCounter, MultiCounter, BreadthFirstIterator



def integrate_singleblock(num_live_points, pointstore, x_dim, num_params, dlogz=0.5):
	active_u = []
	active_v = []
	active_logl = []
	for i in range(num_live_points):
		idx, row = pointstore.pop(-np.inf)
		assert row is not None
		active_u.append(row[2:2+x_dim])
		active_v.append(row[2+x_dim:2+x_dim+num_params])
		active_logl.append(row[1])

	saved_v = []  # Stored points for posterior results
	saved_logl = []
	saved_logwt = []
	h = 0.0  # Information, initially 0.
	logz = -1e300  # ln(Evidence Z), initially Z=0
	logvol = 0
	logvolf = np.log1p(- np.exp(-1.0 / num_live_points))
	#fraction_remain = 1.0
	max_iters = 10000000

	for it in range(0, max_iters):

		# Worst object in collection and its weight (= volume * likelihood)
		worst = np.argmin(active_logl)
		logwt = logvol + logvolf + active_logl[worst]

		# Update evidence Z and information h.
		logz_new = np.logaddexp(logz, logwt)
		h = (np.exp(logwt - logz_new) * active_logl[worst] + np.exp(logz - logz_new) * (h + logz) - logz_new)
		logz = logz_new
		logz_remain = np.max(active_logl) - it / num_live_points

		#print("L=%.1f N=%d V=%.2e logw=%.2e logZ=%.1f logZremain=%.1f" % (active_logl[worst], num_live_points, logvol, logwt, logz, logz_remain))
		# Shrink interval
		logvol -= 1.0 / num_live_points

		# Add worst object to samples.
		saved_v.append(np.array(active_v[worst]))
		saved_logwt.append(logwt)
		saved_logl.append(active_logl[worst])

		# The new likelihood constraint is that of the worst object.
		loglstar = active_logl[worst]

		idx, row = pointstore.pop(loglstar)
		assert row is not None
		u = row[2:2+x_dim]
		v = row[2+x_dim:2+x_dim+num_params]
		logl = row[1]
		
		active_u[worst] = u
		active_v[worst] = v
		active_logl[worst] = logl

		#fraction_remain = np.logaddexp(logz, logz_remain) - logz

		# Stopping criterion
		if logz_remain < logz:
			break
	
	logvol = -len(saved_v) / num_live_points
	for i in np.argsort(active_logl):
		logwt = logvol - np.log(num_live_points) + active_logl[i]
		logz_new = np.logaddexp(logz, logwt)
		h = (np.exp(logwt - logz_new) * active_logl[i] + np.exp(logz - logz_new) * (h + logz) - logz_new)
		logz = logz_new
		#print("L=%.1f N=%d V=%.2e logw=%.2e logZ=%.1f" % (active_logl[i], num_live_points, logvol, logwt, logz))
		saved_v.append(np.array(active_v[i]))
		saved_logwt.append(logwt)
		saved_logl.append(active_logl[i])

	saved_v = np.array(saved_v)
	saved_wt = np.exp(np.array(saved_logwt) - logz)
	saved_logl = np.array(saved_logl)
	logzerr = np.sqrt(h / num_live_points)

	results = dict(niter=it, logz=logz, logzerr=logzerr,
		weighted_samples=dict(v=saved_v, w = saved_wt, logw = saved_logwt, L=saved_logl),
	)
	
	return results


def strategy_advice(node, parallel_values, main_iterator, counting_iterators, rootid):
	if len(node.children) > 0:
		# we don't expand if node already has children
		print("not expanding, already has children")
		assert False
		return np.nan, np.nan
	
	Lmin = parallel_values.min()
	Lmax = parallel_values.max()
	
	logZremain = main_iterator.logZremain
	
	# if the remainder dominates, return that range
	if logZremain > main_iterator.logZ:
		return Lmin, Lmax
	
	#print("not expanding, remainder not dominant")
	return np.nan, np.nan

class __Point(object):
	def __init__(self, u, p):
		self.u = u
		self.p = p


def integrate_graph_singleblock(num_live_points, pointstore, x_dim, num_params, dlogz=0.5):
	pp = PointPile(x_dim, num_params)
	def create_node(pointstore, Lmin):
		idx, row = pointstore.pop(Lmin)
		assert row is not None
		L = row[1]
		u = row[2:2+x_dim]
		p = row[2+x_dim:2+x_dim+num_params]
		assert np.isfinite(L)
		return pp.make_node(L, u, p)
	
	# we create a bunch of live points from the prior volume
	# each of which is the start of a chord (in the simplest case)
	roots = [create_node(pointstore, -np.inf) for i in range(num_live_points)]
	
	iterator_roots = []
	np.random.seed(1)
	for i in range(10):
		# boot-strap which roots are assigned to this iterator
		rootids = np.unique(np.random.randint(len(roots), size=len(roots)))
		#print(rootids)
		iterator_roots.append((SingleCounter(random=True), rootids))
	
	# and we have one that operators on the entire tree
	main_iterator = SingleCounter()
	main_iterator.Lmax = max(n.value for n in roots)
	assert np.isfinite(main_iterator.Lmax)
	
	explorer = BreadthFirstIterator(roots)
	Llo, Lhi = -np.inf, np.inf
	strategy_stale = True
	
	saved_nodeids = []
	saved_logl = []
	
	# we go through each live point (regardless of root) by likelihood value
	while True:
		#print()
		next = explorer.next_node()
		if next is None:
			break
		rootid, node, (active_nodes, active_rootids, active_values, active_node_ids) = next
		# this is the likelihood level we have to improve upon
		Lmin = node.value
		
		saved_nodeids.append(node.id)
		saved_logl.append(Lmin)
		
		expand_node = Lmin <= Lhi and Llo <= Lhi
		# if within suggested range, expand
		if strategy_stale or not (Lmin <= Lhi):
			# check with advisor if we want to expand this node
			Llo, Lhi = strategy_advice(node, active_values, main_iterator, [], rootid)
			#print("L range to expand:", Llo, Lhi, "have:", Lmin, "=>", Lmin <= Lhi, Llo <= Lhi)
			strategy_stale = False
		strategy_stale = True
		
		if expand_node:
			# sample a new point above Lmin
			#print("replacing node", Lmin, "from", rootid, "with", L)
			node.children.append(create_node(pointstore, Lmin))
			main_iterator.Lmax = max(main_iterator.Lmax, node.children[0].value)
		else:
			#print("ending node", Lmin)
			pass
		
		# inform iterators (if it is their business) about the arc
		main_iterator.passing_node(node, active_values)
		for it, rootids in iterator_roots:
			if rootid in rootids:
				mask = np.in1d(active_rootids, rootids, assume_unique=True)
				#mask1 = np.array([rootid2 in rootids for rootid2 in active_rootids])
				#assert (mask1 == mask).all(), (mask1, mask)
				it.passing_node(node, active_values[mask])
		#print([it.H for it,_ in iterator_roots])
		
		explorer.expand_children_of(rootid, node)
		
	# points with weights
	#saved_u = np.array([pp[nodeid].u for nodeid in saved_nodeids])
	saved_v = pp.getp(saved_nodeids)
	saved_logwt = np.array(main_iterator.logweights)
	saved_wt = np.exp(saved_logwt - main_iterator.logZ)
	saved_logl = np.array(saved_logl)
	print('%.4f +- %.4f (main)' % (main_iterator.logZ, main_iterator.logZerr))
	Zest = np.array([it.logZ for it, _ in iterator_roots])
	print('%.4f +- %.4f (bs)' % (Zest.mean(), Zest.std()))

	results = dict(niter=len(saved_logwt), 
		logz=main_iterator.logZ, logzerr=main_iterator.logZerr,
		weighted_samples=dict(v=saved_v, w = saved_wt, logw = saved_logwt, L=saved_logl),
		tree=TreeNode(-np.inf, children=roots),
	)
	
	# return entire tree
	return results


def multi_integrate_graph_singleblock(num_live_points, pointstore, x_dim, num_params, dlogz=0.5, withtests=False):
	pp = PointPile(x_dim, num_params)
	def create_node(pointstore, Lmin):
		idx, row = pointstore.pop(Lmin)
		assert row is not None
		L = row[1]
		u = row[2:2+x_dim]
		p = row[2+x_dim:2+x_dim+num_params]
		return pp.make_node(L, u, p)
	
	# we create a bunch of live points from the prior volume
	# each of which is the start of a chord (in the simplest case)
	roots = [create_node(pointstore, -np.inf) for i in range(num_live_points)]
	
	# and we have one that operators on the entire tree
	main_iterator = MultiCounter(nroots=len(roots), nbootstraps=10, random=True, check_insertion_order=withtests)
	main_iterator.Lmax = max(n.value for n in roots)
	
	explorer = BreadthFirstIterator(roots)
	Llo, Lhi = -np.inf, np.inf
	strategy_stale = True
	
	saved_nodeids = []
	saved_logl = []
	
	# we go through each live point (regardless of root) by likelihood value
	while True:
		#print()
		next = explorer.next_node()
		if next is None:
			break
		rootid, node, (active_nodes, active_rootids, active_values, active_nodeids) = next
		assert not isinstance(rootid, float)
		# this is the likelihood level we have to improve upon
		Lmin = node.value
		
		saved_nodeids.append(node.id)
		saved_logl.append(Lmin)
		
		expand_node = Lmin <= Lhi and Llo <= Lhi
		# if within suggested range, expand
		if strategy_stale or not (Lmin <= Lhi):
			# check with advisor if we want to expand this node
			Llo, Lhi = strategy_advice(node, active_values, main_iterator, [], rootid)
			#print("L range to expand:", Llo, Lhi, "have:", Lmin, "=>", Lmin <= Lhi, Llo <= Lhi)
			strategy_stale = False
		strategy_stale = True
		
		if expand_node:
			# sample a new point above Lmin
			node.children.append(create_node(pointstore, Lmin))
			main_iterator.Lmax = max(main_iterator.Lmax, node.children[0].value)
		else:
			#print("ending node", Lmin)
			pass
		
		# inform iterators (if it is their business) about the arc
		
		assert not isinstance(rootid, float)
		main_iterator.passing_node(rootid, node, active_rootids, active_values)
		
		explorer.expand_children_of(rootid, node)

	print('tree size:', count_tree(roots))
		
	# points with weights
	#saved_u = pp.getu(saved_nodeids)
	saved_v = pp.getp(saved_nodeids)
	saved_logwt = np.array(main_iterator.logweights)
	saved_wt = np.exp(saved_logwt - main_iterator.logZ)
	saved_logl = np.array(saved_logl)
	print('%.4f +- %.4f (main)' % (main_iterator.logZ, main_iterator.logZerr))
	print('%.4f +- %.4f (bs)' % (main_iterator.all_logZ[1:].mean(), main_iterator.all_logZ[1:].std()))
	if withtests:
		print("insertion order:", float(main_iterator.insertion_order_runlength))

	results = dict(niter=len(saved_logwt), 
		logz=main_iterator.logZ, logzerr=main_iterator.logZerr,
		weighted_samples=dict(v=saved_v, w = saved_wt, logw = saved_logwt, L=saved_logl),
		tree=TreeNode(-np.inf, children=roots),
	)
	
	# return entire tree
	return results

testfile = os.path.join(os.path.dirname(__file__), 'eggboxpoints.tsv')

import time
import pytest


@pytest.mark.parametrize("nlive", [100])
def test_singleblock(nlive):
	assert os.path.exists(testfile), ("%s does not exist" % testfile)
	print("="*80)
	print("NLIVE=%d " % nlive)
	print("Standard integrator")
	pointstore = TextPointStore(testfile, 2 + 2 + 2)
	t = time.time()
	result = integrate_singleblock(num_live_points=nlive, pointstore=pointstore, num_params=2, x_dim=2)
	print('  %(logz).1f +- %(logzerr).1f in %(niter)d iter' % result, '%.2fs' % (time.time() - t))
	pointstore.close()
	
	print("Graph integrator")
	pointstore = TextPointStore(testfile, 2 + 2 + 2)
	t = time.time()
	result2 = integrate_graph_singleblock(num_live_points=nlive, pointstore=pointstore, num_params=2, x_dim=2)
	print('  %(logz).1f +- %(logzerr).1f in %(niter)d iter' % result2, '%.2fs' % (time.time() - t))
	pointstore.close()
	assert np.isclose(result2['logz'], result['logz'])
	
	print("Vectorized graph integrator")
	pointstore = TextPointStore(testfile, 2 + 2 + 2)
	t = time.time()
	result3 = multi_integrate_graph_singleblock(num_live_points=nlive, pointstore=pointstore, num_params=2, x_dim=2)
	print('  %(logz).1f +- %(logzerr).1f in %(niter)d iter' % result3, '%.2fs' % (time.time() - t))
	pointstore.close()
	assert np.isclose(result3['logz'], result['logz'])

	print("Vectorized graph integrator with insertion order test")
	pointstore = TextPointStore(testfile, 2 + 2 + 2)
	t = time.time()
	result3 = multi_integrate_graph_singleblock(num_live_points=nlive, pointstore=pointstore, num_params=2, x_dim=2, withtests=True)
	print('  %(logz).1f +- %(logzerr).1f in %(niter)d iter' % result3, '%.2fs' % (time.time() - t))
	pointstore.close()
	assert np.isclose(result3['logz'], result['logz'])

def test_visualisation():
	print("testing tree visualisation...")
	pp = PointPile(1, 1)
	tree = TreeNode()
	for i in range(5):
		j = np.random.randint(1000)
		node = pp.make_node(j, np.array([j]), np.array([j]))
		for k in range(i):
			j = np.random.randint(1000)
			node2 = pp.make_node(j, [j], [j])
			node.children.append(node2)
		tree.children.append(node)
	print(tree)
	print_tree(tree.children, title='Empty Tree')
	

def test_treedump():
	print("testing tree dumping...")
	pp = PointPile(1, 1)
	tree = TreeNode()
	for i in range(5):
		j = np.random.randint(1000)
		node = pp.make_node(j, np.array([j]), np.array([j]))
		for k in range(i):
			j = np.random.randint(1000)
			node2 = pp.make_node(j, [j], [j])
			node.children.append(node2)
		tree.children.append(node)
	dump_tree("test_tree.hdf5", tree.children, pp)
	os.remove("test_tree.hdf5")
	dump_tree("test_tree.hdf5", roots=tree.children, pointpile=pp)
	dump_tree("test_tree.hdf5", tree.children, pp)
	os.remove("test_tree.hdf5")
	

if __name__ == '__main__':
	for nlive in [100, 400, 2000]:
		test_singleblock(nlive)
	#pointstore = TextPointStore(testfile, 2 + 2 + 2)
	#nlive = 400
	#multi_integrate_graph_singleblock(num_live_points=nlive, pointstore=pointstore, num_params=2, x_dim=2)
	#pointstore.close()



from ultranest.mlfriends import MLFriends, ScalingLayer, AffineLayer
import numpy as np
import time
import matplotlib.pyplot as plt

def benchmark_maxradius():
	print(" ndim |  npts | duration")
	for ndim in 2, 4, 8, 16, 32, 64:
		plotpoints = []
		np.random.seed(ndim)
		for npts in 100, 400, 1000, 4000:
			points = np.random.uniform(size=(npts,ndim))
			transformLayer = ScalingLayer()
			region = MLFriends(points, transformLayer)
			
			niter = 0
			total_duration = 0
			while total_duration < 1:
				start = time.time()
				maxr = region.compute_maxradiussq(nbootstraps=20)
				total_duration += time.time() - start
				niter += 1
			print('%5d | %5d | %.2fms  val=%f' % (ndim, npts, total_duration * 1000 / niter, maxr))
			plotpoints.append((npts, total_duration * 1000 / niter / npts**2))
		plt.plot(*zip(*plotpoints), label='ndim=%d' % ndim)
	
	plt.xlabel('Number of live points')
	plt.ylabel('Duration [ms] / nlive$^2$')
	plt.yscale('log')
	plt.xscale('log')
	plt.legend(loc='best', prop=dict(size=10))
	plt.savefig('testmaxradius.pdf', bbox_inches='tight')
	plt.close()
	

def benchmark_transform():
	npts = 400
	for layer in 'scale', 'affine':
		print(" ndim | duration  [%s]" % layer)
		tplotpoints = []
		rplotpoints = []
		nplotpoints = []
		for ndim in 2, 4, 8, 16, 32, 64, 128, 256,:
			np.random.seed(ndim)
			points = np.random.uniform(0.4, 0.6, size=(npts,ndim))
			transformLayer = ScalingLayer() if layer == 'scale' else AffineLayer() 
			region = MLFriends(points, transformLayer)
			region.maxradiussq, region.enlarge = region.compute_enlargement(nbootstraps=30)
			region.create_ellipsoid()
			
			niter = 0
			total_duration = 0
			while total_duration < .1:
				start = time.time()
				u = region.transformLayer.untransform(np.random.normal(size=(ndim)))
				region.transformLayer.transform(u)
				total_duration += time.time() - start
				niter += 1
			print('%5d | %.2fms ' % (ndim, total_duration * 1000 / niter))
			tplotpoints.append((ndim, total_duration * 1000 / niter))

			niter = 0
			total_duration = 0
			while total_duration < .1:
				u = np.random.normal(0.5, 0.1, size=(10, ndim))
				start = time.time()
				region.inside(u)
				total_duration += time.time() - start
				niter += 1
			print('%5d |          %.2fms ' % (ndim, total_duration * 1000 / niter))
			rplotpoints.append((ndim, total_duration * 1000 / niter))
			
			niter = 0
			total_duration = 0
			while total_duration < .1:
				u = np.random.normal(0.5, 0.1, size=(10, ndim))
				start = time.time()
				array = np.empty((10), dtype=int)
				array[:] = -1
				array = np.empty((10), dtype=int)
				array[:] = -1
				array = np.empty((10), dtype=int)
				array[:] = -1
				total_duration += time.time() - start
				niter += 1
			print('%5d |                 %.2fms ' % (ndim, total_duration * 1000 / niter))
			nplotpoints.append((ndim, total_duration * 1000 / niter))
		plt.plot(*zip(*tplotpoints), label=layer + ' transform')
		plt.plot(*zip(*rplotpoints), label=layer + ' region.inside')
		plt.plot(*zip(*nplotpoints), label=layer + ' array')
	
	plt.xlabel('Number of dimensions')
	plt.ylabel('Duration [ms]')
	plt.yscale('log')
	plt.xscale('log')
	plt.legend(loc='best', prop=dict(size=10))
	plt.savefig('testtransform.pdf', bbox_inches='tight')
	plt.close()
		
if __name__ == '__main__':
	#benchmark_maxradius()
	benchmark_transform()



import numpy as np
from ultranest import ReactiveNestedSampler
from ultranest.popstepsampler import PopulationSliceSampler, generate_cube_oriented_direction, \
    generate_random_direction, generate_region_oriented_direction, generate_region_random_direction

def loglike_vectorized(z):
    a = np.array([-0.5 * sum([((xi - 0.7 + i*0.001)/0.1)**2 for i, xi in enumerate(x)]) for x in z])
    b = np.array([-0.5 * sum([((xi - 0.3 - i*0.001)/0.1)**2 for i, xi in enumerate(x)]) for x in z])
    return np.logaddexp(a, b)

def loglike(x):
    a = -0.5 * sum([((xi - 0.7 + i*0.001)/0.1)**2 for i, xi in enumerate(x)])
    b = -0.5 * sum([((xi - 0.3 - i*0.001)/0.1)**2 for i, xi in enumerate(x)])
    return np.logaddexp(a, b)

def transform(x):
    return x # * 10. - 5.

paramnames = ['param%d' % i for i in range(3)]

def test_stepsampler_cubeslice(plot=False):
    np.random.seed(3)
    nsteps = np.random.randint(10, 50)
    popsize = np.random.randint(1, 20)
    sampler = ReactiveNestedSampler(paramnames, loglike_vectorized, transform=transform, vectorized=True)

    sampler.stepsampler = PopulationSliceSampler(
        popsize=popsize, nsteps=nsteps, 
        generate_direction=generate_cube_oriented_direction,
    )
    r = sampler.run(viz_callback=None, log_interval=50)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1
    assert b.sum() > 1

from ultranest.mlfriends import update_clusters, AffineLayer, ScalingLayer, MLFriends, RobustEllipsoidRegion, SimpleRegion

def test_direction_proposals():
    proposals = [generate_cube_oriented_direction, generate_random_direction, 
        generate_region_oriented_direction, generate_region_random_direction]

    points = np.random.uniform(size=(100, 10))
    minvol = 1.0

    scale = 1. # np.random.uniform()
    for layer in AffineLayer, ScalingLayer:
        transformLayer = layer()
        transformLayer.optimize(points, points)
        for region_class in MLFriends, RobustEllipsoidRegion, SimpleRegion:
            region = region_class(points, transformLayer)
            r, f = region.compute_enlargement(minvol=minvol, nbootstraps=30)
            region.maxradiussq = r
            region.enlarge = f
            region.create_ellipsoid(minvol=minvol)

            for prop in proposals:
                print("test of proposal:", prop, "with region:", region_class, "layer:", layer)
                directions = prop(points, region, scale=scale)
                norms = np.linalg.norm(directions, axis=1)
                #print(norms[0], directions[0])
                assert directions.shape == points.shape, (directions.shape, points.shape)
                #assert np.allclose(norms, scale), (norms, scale)

if __name__ == '__main__':
    test_direction_proposals()



# cython: language_level=3,annotate=True,profile=True,fast_fail=True,warning_errors=True
"""Construct and sample from region.

Implements MLFriends efficiently, with transformation layers and clustering.
"""

import numpy as np
cimport numpy as np
from numpy import pi
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
cdef count_nearby(
    np.ndarray[np.float_t, ndim=2] apts,
    np.ndarray[np.float_t, ndim=2] bpts,
    np.float_t radiussq,
    np.ndarray[np.int_t, ndim=1] nnearby
):
    """Count the number of points in ``apts`` within square radius ``radiussq`` for each point ``b`` in `bpts``.

    The number is written to ``nnearby`` (of same length as ``bpts``).

    Parameters
    ----------
    apts: array
        points
    bpts: array
        points
    radiussq: float
        square of the MLFriends radius
    nnearby: array of ints
        The result will be written here.
    """
    cdef size_t na = apts.shape[0]
    cdef size_t nb = bpts.shape[0]
    cdef size_t ndim = apts.shape[1]

    cdef unsigned long i, j
    cdef np.float_t d

    # go through the unselected points and find the worst case
    for j in range(nb):
        # find the nearest selected point
        nnearby[j] = 0
        for i in range(na):
            d = 0
            for k in range(ndim):
                d += (apts[i,k] - bpts[j,k])**2
            if d <= radiussq:
                nnearby[j] += 1


@cython.boundscheck(False)
@cython.wraparound(False)
def find_nearby(
    np.ndarray[np.float_t, ndim=2] apts,
    np.ndarray[np.float_t, ndim=2] bpts,
    np.float_t radiussq,
    np.ndarray[np.int_t, ndim=1] nnearby
):
    """Gets the index of a point in `a` within square radius `radiussq`, for each point `b` in `bpts`.

    The number is written to `nnearby` (of same length as `bpts`).
    If none is found, -1 is written.

    Parameters
    ----------
    apts: array
        points
    bpts: array
        points
    radiussq: float
        square of the MLFriends radius
    nnearby: array of ints
        The result will be written here.

    """
    cdef size_t na = apts.shape[0]
    cdef size_t nb = bpts.shape[0]
    cdef size_t ndim = apts.shape[1]

    cdef unsigned long i, j
    cdef np.float_t d

    # go through the unselected points and find the worst case
    for j in range(nb):
        # find the nearest selected point
        nnearby[j] = -1
        for i in range(na):
            d = 0.0
            for k in range(ndim):
                d += (apts[i,k] - bpts[j,k])**2
            if d <= radiussq:
                nnearby[j] = i
                break


@cython.boundscheck(False)
@cython.wraparound(False)
cdef float compute_maxradiussq(np.ndarray[np.float_t, ndim=2] apts, np.ndarray[np.float_t, ndim=2] bpts):
    """Measure the euclidean distance to the nearest point in `apts`, for each point `b` in `bpts`.

    Parameters
    ----------
    apts: array
        points
    bpts: array
        points

    Returns
    ---------
    array containing the square euclidean distance for each point in bpts.
    """
    cdef size_t na = apts.shape[0]
    cdef size_t nb = bpts.shape[0]
    cdef size_t ndim = apts.shape[1]

    cdef unsigned long i, j
    cdef np.float_t d
    cdef np.float_t mind = 1e300
    cdef np.float_t maxd = 0

    # go through the unselected points and find the worst case
    for j in range(nb):

        # find the nearest selected point
        mind = 1e300
        for i in range(na):
            d = 0
            for k in range(ndim):
                d += (apts[i,k] - bpts[j,k])**2
            mind = min(mind, d)

        maxd = max(maxd, mind)

    return maxd


@cython.boundscheck(False)
@cython.wraparound(False)
def compute_mean_pair_distance(
    np.ndarray[np.float_t, ndim=2] pts,
    np.ndarray[np.int_t, ndim=1] clusterids
):
    """Compute the average distance between pairs of points.
    Pairs from different clusters are excluded in the computation.

    Parameters
    ----------
    pts: array
        points
    clusterids: array of ints or None
        for each point, index of the associated cluster.

    Returns
    ---------
    mean distance between point pairs.
    """
    cdef size_t na = pts.shape[0]
    cdef size_t ndim = pts.shape[1]

    cdef unsigned long i, j
    cdef np.float_t total_dist = 0.0
    cdef np.float_t pair_dist
    cdef int Npairs = 0

    # go through the unselected points and find the worst case
    for j in range(na):
        if clusterids[j] == 0:
            continue
        # find the nearest selected point
        for i in range(j):
            # only consider points in the same cluster
            if clusterids[j] == clusterids[i]:
                pair_dist = 0.0
                for k in range(ndim):
                    pair_dist += (pts[i,k] - pts[j,k])**2
                total_dist += pair_dist**0.5
                Npairs += 1

    assert np.isfinite(total_dist), total_dist
    return total_dist / Npairs


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _update_clusters(
    np.ndarray[np.float_t, ndim=2] upoints,
    np.ndarray[np.float_t, ndim=2] tpoints,
    np.float_t maxradiussq,
    np.ndarray[np.int_t, ndim=1] clusterids,
):
    """same signature as ``update_clusters()``, see there."""
    assert upoints.shape[0] == tpoints.shape[0], ('different number of points', upoints.shape[0], tpoints.shape[0])
    assert upoints.shape[1] == tpoints.shape[1], ('different dimensionality of points', upoints.shape[1], tpoints.shape[1])
    clusteridxs = np.zeros(len(tpoints), dtype=int)
    currentclusterid = 1
    i = 0
    # avoid issues when old clusterids are from a longer array
    clusterids = clusterids[:len(tpoints)]
    existing = clusterids == currentclusterid
    if existing.any():
        i = np.where(existing)[0][0]

    clusteridxs[i] = currentclusterid
    while True:
        # compare known members to unassociated
        nonmembermask = clusteridxs == 0
        if not nonmembermask.any():
            # everyone has been assigned -> done!
            break

        nonmembers = tpoints[nonmembermask,:]
        idnearby = np.empty(len(nonmembers), dtype=int)
        members = tpoints[clusteridxs == currentclusterid,:]
        find_nearby(members, nonmembers, maxradiussq, idnearby)
        # print('merging %d into cluster %d of size %d' % (np.count_nonzero(nnearby), currentclusterid, len(members)))

        if (idnearby >= 0).any():
            # place into cluster
            newmembers = nonmembermask
            newmembers[nonmembermask] = idnearby >= 0
            # print('adding', newmembers.sum())
            clusteridxs[newmembers] = currentclusterid
        else:
            # start a new cluster
            currentclusterid = currentclusterid + 1
            i = np.where(nonmembermask)[0][0]
            if clusterids is not None:
                existing = clusterids == currentclusterid
                if existing.any():
                    i = np.where(existing)[0][0]

            clusteridxs[i] = currentclusterid

    assert (clusteridxs > 0).all()
    nclusters = len(np.unique(clusteridxs))
    # assert np.all(np.unique(clusteridxs) == np.arange(nclusters)+1), (np.unique(clusteridxs), nclusters, np.arange(nclusters)+1)
    if nclusters == 1:
        overlapped_upoints = upoints
    else:
        overlapped_upoints = np.empty_like(upoints)
        for idx in np.unique(clusteridxs):
            group_upoints = upoints[clusteridxs == idx,:]
            if len(group_upoints) > 1:
                # center on group mean
                group_mean = group_upoints.mean(axis=0).reshape((1,-1))
            else:
                # if a single point, the differences would be zero,
                # not giving weight to this being an outlier.
                # so use the mean of the entire point population instead
                group_mean = upoints.mean(axis=0).reshape((1,-1))
            overlapped_upoints[clusteridxs == idx,:] = group_upoints - group_mean

    return nclusters, clusteridxs, overlapped_upoints


@cython.boundscheck(False)
@cython.wraparound(False)
def update_clusters(
    np.ndarray[np.float_t, ndim=2] upoints,
    np.ndarray[np.float_t, ndim=2] tpoints,
    np.float_t maxradiussq,
    clusterids=None,
):
    """Clusters `upoints`, so that clusters are distinct if no 
    member pair is within a radius of sqrt(`maxradiussq`).

    Parameters
    ----------
    upoints: array
        points (in u-space)
    tpoints: array
        points (in t-space)
    maxradiussq: float
        square of the MLFriends radius
    clusterids: array of ints or None
        for each point, index of the associated cluster.

    Returns
    ---------
    nclusters: int
        the number of clusters found, which is also clusterids.max()
    new_clusterids: array of int
        the new clusterids for each point
    overlapped_points:
        upoints with their cluster centers subtracted.

    The existing cluster ids are re-used when assigning new clusters,
    if possible.
    Clustering is performed on a transformed coordinate space (`tpoints`).
    Returned values are based on upoints.
    """
    if clusterids is None:
        clusterids = np.zeros(len(tpoints), dtype=int)
    return _update_clusters(upoints, tpoints, maxradiussq, clusterids)


@cython.boundscheck(False)
@cython.wraparound(False)
def make_eigvals_positive(
    np.ndarray[np.float_t, ndim=2] a,
    np.float_t targetprod
):
    """For the symmetric square matrix ``a``, increase any zero eigenvalues
    to fulfill a target product of eigenvalues.

    Parameters
    ----------
    a: array
        covariance matrix
    targetprod: array
        target product of eigenvalues

    Returns
    ---------
    covariance matrix
    """

    assert np.isfinite(a).all(), a
    try:
        w, v = np.linalg.eigh(a)  # Use eigh because we assume a is symmetric.
    except np.linalg.LinAlgError as e:
        print(a, targetprod)
        raise e
    mask = w < max(1.e-10, 1e-300**(1. / len(a)))
    if np.any(mask):
        nzprod = np.product(w[~mask])  # product of nonzero eigenvalues
        nzeros = mask.sum()  # number of zero eigenvalues
        w[mask] = (targetprod / nzprod) ** (1. / nzeros)  # adjust zero eigvals
        a = np.dot(np.dot(v, np.diag(w)), np.linalg.inv(v))  # re-form cov

    return a


@cython.boundscheck(False)
@cython.wraparound(False)
def bounding_ellipsoid(
    np.ndarray[np.float_t, ndim=2] x,
    np.float_t minvol=0.
):
    """Calculate bounding ellipsoid containing a set of points x.

    Parameters
    ----------
    x : (npoints, ndim) ndarray
        Coordinates of uniformly sampled points.
    pointvol : float, optional
        Used to set a minimum bound on the ellipsoid volume when
        minvol is True.

    Returns
    -------
    mean and covariance of points

    """
    # Function taken from nestle, MIT licensed, (C) kbarbary

    ndim = x.shape[1]

    # Calculate covariance of points
    ctr = np.mean(x, axis=0)
    delta = x - ctr
    cov = np.cov(delta, rowvar=0)
    assert np.isfinite(cov).all(), (cov, x)
    if ndim == 1:
        cov = np.atleast_2d(cov)

    # For a ball of uniformly distributed points, the covariance will be
    # smaller than r^2 by a factor of 1/(n+2) [see, e.g.,
    # http://mathoverflow.net/questions/35276/
    # covariance-of-points-distributed-in-a-n-ball]. In nested sampling,
    # we are supposing the points are uniformly distributed within
    # an ellipse, so the same factor holds. Expand `cov`
    # to compensate for that when defining the ellipse matrix:
    cov *= (ndim + 2)

    # Ensure that ``cov`` is nonsingular.
    # It can be singular when the ellipsoid has zero volume, which happens
    # when npoints <= ndim or when enough points are linear combinations
    # of other points. (e.g., npoints = ndim+1 but one point is a linear
    # combination of others). When this happens, we expand the ellipse
    # in the zero dimensions to fulfill the volume expected from
    # ``pointvol``.
    if minvol > 0:
        cov = make_eigvals_positive(cov, minvol)

    return ctr, cov


class ScalingLayer(object):
    """Simple transformation layer that only shifts and scales each axis."""

    def __init__(self, mean=0, std=1, nclusters=1, wrapped_dims=[], clusterids=None):
        """Initialise layer."""
        self.mean = mean
        self.std = std
        self.nclusters = nclusters
        self.wrapped_dims = wrapped_dims
        self.has_wraps = len(wrapped_dims) > 0
        self.clusterids = clusterids

    def optimize_wrap(self, points):
        """Optimization for wrapped/circular parameters.
        Does nothing if there are no wrapped/circular parameters.

        Parameters
        ----------
        points: array
            points to use for optimization (in u-space)

        Find largest gap in live points and wrap parameter space there.

        For example, if a wrapped axis has::

            |*****           ****|

        it would identify the middle, subtract it, so that the new space
        is::

            |      ********      |
        """
        if not self.has_wraps:
            return 

        N, ndims = points.shape
        self.wrap_cuts = []
        for i in self.wrapped_dims:
            # find largest gap
            vals = np.pad(points[:,i], 1, mode='constant', constant_values=(0,1))
            vals.sort()
            assert vals[0] == 0
            assert vals[-1] == 1
            deltas = vals[1:] - vals[:-1]
            j = deltas.argmax()

            # wrap between i, i+1
            cut = (vals[j] + vals[j+1]) / 2.
            self.wrap_cuts.append(cut)

    def wrap(self, points):
        """Wrap points for circular parameters."""
        if not self.has_wraps:
            return points
        wpoints = points.copy().reshape((-1, points.shape[-1]))
        for i, cut in zip(self.wrapped_dims, self.wrap_cuts):
            wpoints[:,i] = np.fmod(wpoints[:,i] + (1 - cut), 1)
        return wpoints

    def unwrap(self, wpoints):
        """Undo wrapping for circular parameters."""
        if not self.has_wraps:
            return wpoints
        points = wpoints.copy().reshape((-1, wpoints.shape[-1]))
        for i, cut in zip(self.wrapped_dims, self.wrap_cuts):
            points[:,i] = np.fmod(points[:,i] + cut, 1)
        return points

    def optimize(self, points, centered_points, clusterids=None, minvol=0.):
        """Optimize layer.

        Estimates mean and std of points.

        Parameters
        ----------
        points: array
            points to use for optimize (in u-space)
        centered_points: array
            points with their cluster center subtracted
        clusterids: array of ints
            for each point, which cluster they belong to
        minvol:
            ignored
        """
        self.optimize_wrap(points)
        wrapped_points = self.wrap(points)
        self.mean = wrapped_points.mean(axis=0).reshape((1,-1))
        self.std = centered_points.std(axis=0).reshape((1,-1))
        self.axes = np.diag(self.std[0])
        self.logvolscale = np.sum(np.log(self.std))
        self.set_clusterids(clusterids=clusterids, npoints=len(points))

    def set_clusterids(self, clusterids=None, npoints=None):
        """Updates the cluster id assigned to each point."""
        if clusterids is None and self.clusterids is None and npoints is not None:
            # for the beginning, set cluster ids to one for all points
            clusterids = np.ones(npoints, dtype=int)
        if clusterids is not None:
            # if we have a value, update
            self.clusterids = clusterids

    def create_new(self, upoints, maxradiussq, minvol=0.):
        """Learn next layer from this optimized layer's clustering.

        Parameters
        ----------
        upoints: array
            points to use for optimize (in u-space)
        maxradiussq: float
            square of the MLFriends radius
        minvol: float
            Minimum volume to regularize sample covariance

        Returns
        ---------
        A new, optimized ScalingLayer.
        """
        # perform clustering in transformed space
        uwpoints = self.wrap(upoints)
        tpoints = self.transform(upoints)
        nclusters, clusteridxs, overlapped_uwpoints = update_clusters(uwpoints, tpoints, maxradiussq, self.clusterids)
        # clusteridxs = track_clusters(clusteridxs, self.clusterids)
        s = ScalingLayer(nclusters=nclusters, wrapped_dims=self.wrapped_dims, clusterids=clusteridxs)
        s.optimize(upoints, overlapped_uwpoints)
        return s

    def transform(self, u):
        """Transform points from cube space to a whitened space."""
        if self.has_wraps:
            w = self.wrap(u)
        else:
            w = u
        return ((w - self.mean) / self.std).reshape(u.shape)

    def untransform(self, ww):
        """Transform points from whitened space back to cube space."""
        w = (ww * self.std) + self.mean
        if self.has_wraps:
            u = self.unwrap(w).reshape(ww.shape)
        else:
            u = w.reshape(ww.shape)
        return u


class AffineLayer(ScalingLayer):
    """Affine whitening transformation.

    Learns the covariance of points.
    """

    def __init__(self, ctr=0, T=1, invT=1, nclusters=1, wrapped_dims=[], clusterids=None):
        """Initialise layer.

        The parameters are optional and can be learned from points with :meth:`optimize`

        Parameters
        ----------
        ctr: vector
            Center of points
        T: matrix
            transformation matrix
        invT: matrix
            inverse transformation matrix
        nclusters: int
            number of clusters
        wrapped_dims: array of bools
            indicates which parameter axes are circular.
        clusterids: array of int
            cluster id for each point

        """
        self.ctr = ctr
        self.T = T
        self.invT = invT
        self.nclusters = nclusters
        self.wrapped_dims = wrapped_dims
        self.has_wraps = len(wrapped_dims) > 0
        self.clusterids = clusterids

    def optimize(self, points, centered_points, clusterids=None, minvol=0.):
        """Optimize layer.

        Estimates covariance of ``centered_points``. ``minvol`` sets the
        smallest allowed size of the covariance to avoid numerical
        collapse.

        Parameters
        ----------
        points: array
            points to use for optimize (in u-space)
        centered_points: array
            points with their cluster center subtracted
        clusterids: array of ints
            for each point, which cluster they belong to
        minvol: float
            Minimum volume to regularize sample covariance
        """
        self.optimize_wrap(points)
        wrapped_points = self.wrap(points)
        self.ctr = np.mean(wrapped_points, axis=0)
        cov = np.cov(centered_points, rowvar=0)
        cov *= (len(self.ctr) + 2)
        self.cov = cov
        eigval, eigvec = np.linalg.eigh(cov)
        eigvalmin = eigval.max() * 1e-40
        eigval[eigval < eigvalmin] = eigvalmin
        a = np.linalg.inv(cov)
        self.logvolscale = np.linalg.slogdet(a)[1] * -0.5

        self.T = eigvec * eigval**-0.5
        self.invT = np.linalg.inv(self.T)
        self.axes = self.invT
        # print('transform used:', self.T, self.invT, 'cov:', cov, 'eigen:', eigval, eigvec)
        self.set_clusterids(clusterids=clusterids, npoints=len(points))

    def create_new(self, upoints, maxradiussq, minvol=0.):
        """Learn next layer from this optimized layer's clustering.

        Parameters
        ----------
        upoints: array
            points to use for optimize (in u-space)
        maxradiussq: float
            square of the MLFriends radius
        minvol: float
            Minimum volume to regularize sample covariance

        Returns
        ---------
        A new, optimized AffineLayer.
        """
        # perform clustering in transformed space
        uwpoints = self.wrap(upoints)
        tpoints = self.transform(upoints)
        nclusters, clusteridxs, overlapped_uwpoints = update_clusters(uwpoints, tpoints, maxradiussq, self.clusterids)
        # clusteridxs = track_clusters(clusteridxs, self.clusterids)
        s = AffineLayer(nclusters=nclusters, wrapped_dims=self.wrapped_dims, clusterids=clusteridxs)
        s.optimize(upoints, overlapped_uwpoints, minvol=minvol)
        return s

    def transform(self, u):
        """Transform points from cube space to a whitened space."""
        if self.has_wraps:
            w = self.wrap(u)
        else:
            w = u
        return np.dot(w - self.ctr, self.T)

    def untransform(self, ww):
        """Transform points from whitened space back to cube space."""
        w = np.dot(ww, self.invT) + self.ctr
        if self.has_wraps:
            u = self.unwrap(w).reshape(ww.shape)
        else:
            u = w.reshape(ww.shape)
        return u


def vol_prefactor(np.int_t n):
    """Volume constant for an ``n``-dimensional sphere.

    for ``n`` even:  $$    (2pi)^(n    /2) / (2 * 4 * ... * n)$$
    for ``n`` odd :  $$2 * (2pi)^((n-1)/2) / (1 * 3 * ... * n)$$

    Parameters
    ----------
    n: int
        dimensionality

    Returns
    ---------
    volume (float)
    """
    if n % 2 == 0:
        f = 1.
        i = 2
    else:
        f = 2.
        i = 3

    while i <= n:
        f *= 2. / i * pi
        i += 2

    return f


def _inside_ellipsoid(
    np.ndarray[np.float_t, ndim=2] points,
    np.ndarray[np.float_t, ndim=1] ellipsoid_center,
    np.ndarray[np.float_t, ndim=2] ellipsoid_invcov,
    np.float_t square_radius
):
    """Check if inside ellipsoid

    Parameters
    ----------
    points: array of vectors
        Points to check
    ellipsoid_center: vector
        center of ellipsoid
    ellipsoid_invcov: matrix
        inverse covariance matrix
    square_radius: float
        square radius

    Returns
    ---------
    is_inside: array of bools
        True if inside wrapping for each point in ``points``.
    """
    # compute distance vector to center
    d = points - ellipsoid_center
    # distance in normalised coordates: vector . matrix . vector
    # where the matrix is the ellipsoid inverse covariance
    r = np.einsum('ij,jk,ik->i', d, ellipsoid_invcov, d)
    # (r <= 1) means inside
    return r <= square_radius


class MLFriends(object):
    """MLFriends region.

    Defines a region around nested sampling live points for

    1. checking whether a proposed point likely also fulfills the
       likelihood constraints
    2. proposing new points.

    Learns geometry of region from existing live points.
    """

    def __init__(self, u, transformLayer):
        """Initialise region.

        Parameters
        -----------
        u: array of vectors
            live points
        transformLayer: ScalingLayer or AffineLayer
            whitening layer

        """
        if not np.logical_and(u > 0, u < 1).all():
            raise ValueError("not all u values are between 0 and 1: %s" % u[~np.logical_and(u > 0, u < 1).all()])

        self.u = u
        self.set_transformLayer(transformLayer)

        self.sampling_methods = [
            self.sample_from_transformed_boundingbox,
            self.sample_from_boundingbox,
            self.sample_from_points,
            self.sample_from_wrapping_ellipsoid
        ]
        self.current_sampling_method = self.sample_from_boundingbox
        self.vol_prefactor = vol_prefactor(self.u.shape[1])

    def estimate_volume(self):
        """Estimate the order of magnitude of the volume around a single point
        given the current transformLayer.

        Does not account for:
        * the number of live points
        * their overlap
        * the intersection with the unit cube borders

        Returns
        -------
        volume: float
            Volume
        """
        r = self.maxradiussq**0.5
        N, ndim = self.u.shape
        # how large is a sphere of size r in untransformed coordinates?
        return self.transformLayer.logvolscale + np.log(r) * ndim  #+ np.log(vol_prefactor(ndim))

    def set_transformLayer(self, transformLayer):
        """Update transformation layer. Invalidates attribute `maxradius`.

        Parameters
        ----------
        transformLayer: ScalingLayer or AffineLayer
            t-space transformation layer

        """
        self.transformLayer = transformLayer
        self.unormed = self.transformLayer.transform(self.u)
        assert np.isfinite(self.unormed).all(), (self.unormed, self.u)
        self.bbox_lo = self.unormed.min(axis=0)
        self.bbox_hi = self.unormed.max(axis=0)
        self.maxradiussq = None

    def compute_maxradiussq(self, nbootstraps=50):
        """Run MLFriends bootstrapping

        Parameters
        ----------
        nbootstraps: int
            number of bootstrapping rounds

        Returns
        -------
        square radius that safely encloses all live points.
        """
        N, ndim = self.u.shape
        selected = np.empty(N, dtype=bool)
        maxd = 0

        for i in range(nbootstraps):
            idx = np.random.randint(N, size=N)
            selected[:] = False
            selected[idx] = True
            a = self.unormed[selected,:]
            b = self.unormed[~selected,:]

            # compute distances from a to b
            maxd = max(maxd, compute_maxradiussq(a, b))

        assert maxd > 0, (maxd, self.u)
        return maxd

    def compute_enlargement(self, nbootstraps=50, minvol=0., rng=np.random):
        """Return MLFriends radius and ellipsoid enlargement using bootstrapping.

        The wrapping ellipsoid covariance is determined in each bootstrap round.

        Parameters
        ----------
        nbootstraps: int
            number of bootstrapping rounds
        minvol: float
            minimum volume to enforce to wrapping ellipsoid
        rng:
            random number generator

        Returns
        -------
        max_distance: float
            square radius of MLFriends algorithm
        max_radius: float
            square radius of enclosing ellipsoid.
        """
        N, ndim = self.u.shape
        assert np.isfinite(self.unormed).all(), self.unormed
        selected = np.empty(N, dtype=bool)
        maxd = 0.0
        maxf = 0.0

        for i in range(nbootstraps):
            idx = rng.randint(N, size=N)
            selected[:] = False
            selected[idx] = True
            if selected.all() or not selected.any():
                continue

            # compute distances from a to b
            maxd = max(maxd, compute_maxradiussq(
                self.unormed[selected,:],
                self.unormed[~selected,:]))

            # compute enlargement of bounding ellipsoid
            ctr, cov = bounding_ellipsoid(self.u[selected,:], minvol=minvol)
            a = np.linalg.inv(cov)  # inverse covariance
            # compute expansion factor
            delta = self.u[~selected,:] - ctr
            #f = np.einsum('...i, ...i', np.tensordot(delta, a, axes=1), delta).max()
            f = np.einsum('ij,jk,ik->i', delta, a, delta).max()
            assert np.isfinite(f), (ctr, cov, self.unormed, f, delta, a)
            if not f > 0:
                raise np.linalg.LinAlgError("Distances are not positive")
            maxf = max(maxf, f)

        assert maxd > 0, (maxd, self.u, self.unormed)
        assert maxf > 0, (maxf, self.u, self.unormed)
        return maxd, maxf

    def sample_from_points(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Chooses randomly from points and their ellipsoids.

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # generate points near random existing points
        idx = np.random.randint(N, size=nsamples)
        v = np.random.normal(size=(nsamples, ndim))
        v *= (np.random.uniform(size=nsamples)**(1. / ndim) / np.linalg.norm(v, axis=1)).reshape((-1, 1))
        v = self.unormed[idx,:] + v * self.maxradiussq**0.5

        # count how many are around
        nnearby = np.empty(nsamples, dtype=int)
        count_nearby(self.unormed, v, self.maxradiussq, nnearby)
        vmask = np.random.uniform(high=nnearby) < 1
        w = self.transformLayer.untransform(v[vmask,:])
        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        wmask[wmask] = self.inside_ellipsoid(w[wmask])

        return w[wmask,:]

    def sample_from_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region.

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from unit cube in prior space
        u = np.random.uniform(size=(nsamples, ndim))
        wmask = self.inside_ellipsoid(u)
        # check if inside region in transformed space
        v = self.transformLayer.transform(u[wmask,:])
        idnearby = np.empty(len(v), dtype=int)
        find_nearby(self.unormed, v, self.maxradiussq, idnearby)
        vmask = idnearby >= 0
        return u[wmask,:][vmask,:]

    def sample_from_transformed_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region (in whitened space).

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space
        v = np.random.uniform(self.bbox_lo - self.maxradiussq, self.bbox_hi + self.maxradiussq, size=(nsamples, ndim))
        idnearby = np.empty(nsamples, dtype=int)
        find_nearby(self.unormed, v, self.maxradiussq, idnearby)
        vmask = idnearby >= 0

        # check if inside unit cube
        w = self.transformLayer.untransform(v[vmask,:])
        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        wmask[wmask] = self.inside_ellipsoid(w[wmask])

        return w[wmask,:]

    def sample_from_wrapping_ellipsoid(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from wrapping ellipsoid and filters with region.

        Parameters as described in ``sample()``.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space

        z = np.random.normal(size=(nsamples, ndim))
        assert ((z**2).sum(axis=1) > 0).all(), (z**2).sum(axis=1)
        z /= ((z**2).sum(axis=1)**0.5).reshape((nsamples, 1))
        assert self.enlarge > 0, self.enlarge
        u = z * self.enlarge**0.5 * np.random.uniform(size=(nsamples, 1))**(1./ndim)

        w = self.ellipsoid_center + np.dot(u, self.ellipsoid_axes_T)
        #assert self.inside_ellipsoid(w).all()

        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        v = self.transformLayer.transform(w[wmask,:])
        idnearby = np.empty(len(v), dtype=int)
        find_nearby(self.unormed, v, self.maxradiussq, idnearby)
        vmask = idnearby >= 0

        return w[wmask,:][vmask,:]

    def sample(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Switches automatically between the ``sampling_methods`` (attribute).

        Parameters
        ----------
        nsamples: int
            number of samples to draw
        
        Returns
        -------
        samples: array of shape (nsamples, dimension)
            samples drawn
        idx: array of integers (nsamples)
            index of a point nearby (MLFriends.u)
        """
        samples = self.current_sampling_method(nsamples=nsamples)
        if len(samples) == 0:
            # no result, choose another method
            self.current_sampling_method = self.sampling_methods[np.random.randint(len(self.sampling_methods))]
            # print("switching to %s" % self.current_sampling_method)
        return samples

    def inside(self, pts):
        """Check if inside region.

        Parameters
        ----------
        pts: array of vectors
            Points to check

        Returns
        ---------
        is_inside: array of bools
            True if inside MLFriends region and wrapping ellipsoid,
            for each point in ``pts``.

        """
        # require points to be inside bounding ellipsoid
        mask = self.inside_ellipsoid(pts)
        
        if mask.any():
            # additionally require points to be near neighbours
            bpts = self.transformLayer.transform(pts[mask,:])
            idnearby = np.empty(len(bpts), dtype=int)
            find_nearby(self.unormed, bpts, self.maxradiussq, idnearby)
            mask[mask] = idnearby >= 0

        return mask

    def create_ellipsoid(self, minvol=0.0):
        """Create wrapping ellipsoid and store its center and covariance.

        Parameters
        ----------
        minvol: float
            If positive, make sure ellipsoid has at least this volume.
        """
        assert self.enlarge is not None
        # compute enlargement of bounding ellipsoid
        ctr, cov = bounding_ellipsoid(self.u, minvol=minvol)
        a = np.linalg.inv(cov)

        self.ellipsoid_center = ctr
        self.ellipsoid_invcov = a
        self.ellipsoid_cov = cov

        l, v = np.linalg.eigh(a)
        self.ellipsoid_axlens = 1. / np.sqrt(l)
        self.ellipsoid_axes = np.dot(v, np.diag(self.ellipsoid_axlens))
        self.ellipsoid_axes_T = self.ellipsoid_axes.transpose()

        l2, v2 = np.linalg.eigh(cov)
        self.ellipsoid_inv_axlens = 1. / np.sqrt(l2)
        self.ellipsoid_inv_axes = np.dot(v2, np.diag(self.ellipsoid_inv_axlens))


    def inside_ellipsoid(self, u):
        """Check if inside wrapping ellipsoid.

        Parameters
        ----------
        u: array of vectors
            Points to check

        Returns
        ---------
        is_inside: array of bools
            True if inside wrapping ellipsoid, for each point in `pts`.

        """
        return _inside_ellipsoid(u, self.ellipsoid_center, self.ellipsoid_invcov, self.enlarge)

    def compute_mean_pair_distance(self):
        return compute_mean_pair_distance(self.unormed, self.transformLayer.clusterids)


class RobustEllipsoidRegion(MLFriends):
    """Ellipsoidal region.

    Defines a region around nested sampling live points for

    1. checking whether a proposed point likely also fulfills the
       likelihood constraints
    2. proposing new points.

    Learns geometry of region from existing live points.
    """

    def __init__(self, u, transformLayer):
        """Initialise region.

        Parameters
        -----------
        u: array of vectors
            live points
        transformLayer: ScalingLayer or AffineLayer
            whitening layer

        """
        if not np.logical_and(u > 0, u < 1).all():
            raise ValueError("not all u values are between 0 and 1: %s" % u[~np.logical_and(u > 0, u < 1).all()])

        self.u = u
        self.set_transformLayer(transformLayer)

        self.sampling_methods = [
            #self.sample_from_transformed_boundingbox,
            self.sample_from_boundingbox,
            self.sample_from_wrapping_ellipsoid
        ]
        self.current_sampling_method = self.sample_from_boundingbox
        self.vol_prefactor = vol_prefactor(self.u.shape[1])

    def sample_from_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region.

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from unit cube in prior space
        u = np.random.uniform(size=(nsamples, ndim))
        wmask = self.inside_ellipsoid(u)
        return u[wmask,:]

    def sample_from_transformed_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region (in whitened space).

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space
        v = np.random.uniform(self.bbox_lo - self.maxradiussq, self.bbox_hi + self.maxradiussq, size=(nsamples, ndim))

        # check if inside unit cube
        w = self.transformLayer.untransform(v)
        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        wmask[wmask] = self.inside_ellipsoid(w[wmask])

        return w[wmask,:]

    def sample_from_wrapping_ellipsoid(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from wrapping ellipsoid and filters with region.

        Parameters as described in ``sample()``.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space

        z = np.random.normal(size=(nsamples, ndim))
        assert ((z**2).sum(axis=1) > 0).all(), (z**2).sum(axis=1)
        z /= ((z**2).sum(axis=1)**0.5).reshape((nsamples, 1))
        assert self.enlarge > 0, self.enlarge
        u = z * self.enlarge**0.5 * np.random.uniform(size=(nsamples, 1))**(1./ndim)

        w = self.ellipsoid_center + np.dot(u, self.ellipsoid_axes_T)
        #assert self.inside_ellipsoid(w).all()

        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        return w[wmask,:]

    def sample(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Switches automatically between the ``sampling_methods`` (attribute).

        Parameters
        ----------
        nsamples: int
            number of samples to draw
        
        Returns
        -------
        samples: array of shape (nsamples, dimension)
            samples drawn
        idx: array of integers (nsamples)
            index of a point nearby (MLFriends.u)
        """
        samples = self.current_sampling_method(nsamples=nsamples)
        if len(samples) == 0:
            # no result, choose another method
            self.current_sampling_method = self.sampling_methods[np.random.randint(len(self.sampling_methods))]
            #print("switching to %s" % self.current_sampling_method)
        return samples

    def inside(self, pts):
        """Check if inside region.

        Parameters
        ----------
        pts: array of vectors
            Points to check

        Returns
        ---------
        is_inside: array of bools
            True if inside MLFriends region and wrapping ellipsoid,
            for each point in ``pts``.

        """
        # require points to be inside bounding ellipsoid
        return self.inside_ellipsoid(pts)

    def compute_enlargement(self, nbootstraps=50, minvol=0., rng=np.random):
        """Return MLFriends radius and ellipsoid enlargement using bootstrapping.

        The wrapping ellipsoid covariance is determined in each bootstrap round.

        Parameters
        ----------
        nbootstraps: int
            number of bootstrapping rounds
        minvol: float
            minimum volume to enforce to wrapping ellipsoid
        rng:
            random number generator

        Returns
        -------
        max_distance: float
            square radius of MLFriends algorithm
        max_radius: float
            square radius of enclosing ellipsoid.
        """
        N, ndim = self.u.shape
        if N < ndim + 1:
            raise FloatingPointError('not enough live points to compute covariance')
        assert np.isfinite(self.unormed).all(), self.unormed
        selected = np.empty(N, dtype=bool)
        maxd = 1e300
        maxf = 0.0

        for i in range(nbootstraps):
            idx = rng.randint(N, size=N)
            selected[:] = False
            selected[idx] = True

            # compute enlargement of bounding ellipsoid
            ctr, cov = bounding_ellipsoid(self.u[selected,:])
            a = np.linalg.inv(cov)  # inverse covariance
            # compute expansion factor
            delta = self.u[~selected,:] - ctr
            #f = np.einsum('...i, ...i', np.tensordot(delta, a, axes=1), delta).max()
            f = np.einsum('ij,jk,ik->i', delta, a, delta).max()
            assert np.isfinite(f), (ctr, cov, self.unormed, f, delta, a)
            if not f > 0:
                raise np.linalg.LinAlgError("Distances are not positive")
            maxf = max(maxf, f)

        assert maxd > 0, (maxd, self.u, self.unormed)
        assert maxf > 0, (maxf, self.u, self.unormed)
        return maxd, maxf

    def estimate_volume(self):
        """Estimate the volume of the ellipsoid.

        Does not account for the intersection with the unit cube borders.

        Returns
        -------
        logvolume: float
            logarithm of the volume.
        """
        ndim = len(self.ellipsoid_cov)
        sign, logvol = np.linalg.slogdet(self.ellipsoid_cov)
        if sign > 0:
            return logvol + ndim * np.log(self.enlarge)
        else:
            return -1e300


class SimpleRegion(RobustEllipsoidRegion):
    """Axis-aligned ellipsoidal region.

    Defines a region around nested sampling live points for

    1. checking whether a proposed point likely also fulfills the
       likelihood constraints
    2. proposing new points.

    Learns geometry of region from existing live points.
    """

    def create_ellipsoid(self, minvol=0.0):
        """Create wrapping ellipsoid and store its center and covariance.

        Parameters
        ----------
        minvol: float
            If positive, make sure ellipsoid has at least this volume.
        """
        assert self.enlarge is not None
        # compute enlargement of bounding ellipsoid
        ctr = np.mean(self.u, axis=0)
        var = np.var(self.u, axis=0)
        a = np.diag(1. / var)
        cov = np.diag(var)

        self.ellipsoid_center = ctr
        self.ellipsoid_invcov = a
        self.ellipsoid_cov = cov

        l, v = np.linalg.eigh(a)
        self.ellipsoid_axlens = 1. / np.sqrt(l)
        self.ellipsoid_axes = np.dot(v, np.diag(self.ellipsoid_axlens))
        self.ellipsoid_axes_T = self.ellipsoid_axes.transpose()

        l2, v2 = np.linalg.eigh(cov)
        self.ellipsoid_inv_axlens = 1. / np.sqrt(l2)
        self.ellipsoid_inv_axes = np.dot(v2, np.diag(self.ellipsoid_inv_axlens))


    def compute_enlargement(self, nbootstraps=50, minvol=0., rng=np.random):
        """Return MLFriends radius and ellipsoid enlargement using bootstrapping.

        The wrapping ellipsoid covariance is determined in each bootstrap round.

        Parameters
        ----------
        nbootstraps: int
            number of bootstrapping rounds
        minvol: float
            minimum volume to enforce to wrapping ellipsoid
        rng:
            random number generator

        Returns
        -------
        max_distance: float
            square radius of MLFriends algorithm
        max_radius: float
            square radius of enclosing ellipsoid.
        """
        N, ndim = self.u.shape
        assert np.isfinite(self.u).all(), self.u
        assert np.isfinite(self.unormed).all(), self.unormed
        selected = np.empty(N, dtype=bool)
        maxd = 1e300
        maxf = 0.0
        if N < ndim + 1:
            raise FloatingPointError('not enough live points to compute variance')

        for i in range(nbootstraps):
            idx = rng.randint(N, size=N)
            selected[:] = False
            selected[idx] = True

            # compute enlargement of bounding ellipsoid
            ctr = np.mean(self.u[selected,:], axis=0)
            var = np.var(self.u[selected,:], axis=0)
            # compute expansion factor
            f = np.sum((self.u[~selected,:] - ctr.reshape((1, -1)))**2 / var, axis=0).max()
            assert np.isfinite(f), (self.u, ctr, var, self.unormed, f)
            if not f > 0:
                raise np.linalg.LinAlgError("Distances are not positive")
            maxf = max(maxf, f)

        assert maxd > 0, (maxd, self.u, self.unormed)
        assert maxf > 0, (maxf, self.u, self.unormed)
        return maxd, maxf


class WrappingEllipsoid(object):
    """Ellipsoid which safely wraps points."""

    def __init__(self, u):
        """Initialise region.

        Parameters
        -----------
        u: array of vectors
            live points
        """
        self.u = u
        # allow some parameters to have exactly the same value
        # this can occur with grid / categorical parameters
        self.variable_dims = np.std(self.u, axis=0) > 0
        if self.variable_dims.all():
            self.variable_dims = Ellipsis

    def compute_enlargement(self, nbootstraps=50, rng=np.random):
        """Return ellipsoid enlargement after `nbootstraps` bootstrapping rounds.

        The wrapping ellipsoid covariance is determined in each bootstrap round.
        """
        N = len(self.u)
        v = self.u[:,self.variable_dims]
        selected = np.empty(N, dtype=bool)
        maxf = 0.0

        for i in range(nbootstraps):
            idx = rng.randint(N, size=N)
            selected[:] = False
            selected[idx] = True
            ua = v[selected,:]
            ub = v[~selected,:]

            # compute enlargement of bounding ellipsoid
            ctr, cov = bounding_ellipsoid(ua)
            a = np.linalg.inv(cov)  # inverse covariance
            # compute expansion factor
            delta = ub - ctr
            f = np.einsum('...i, ...i', np.tensordot(delta, a, axes=1), delta).max()
            if not f > 0:
                raise np.linalg.LinAlgError("Distances are not positive")
            maxf = max(maxf, f)

        assert maxf > 0, (maxf, self.u, self.active_dims)
        return maxf

    def create_ellipsoid(self, minvol=0.0):
        """Create wrapping ellipsoid and store its center and covariance."""
        assert self.enlarge is not None
        # compute enlargement of bounding ellipsoid
        ctr, cov = bounding_ellipsoid(self.u[:,self.variable_dims], minvol=minvol)
        a = np.linalg.inv(cov)

        self.ellipsoid_center = ctr
        self.ellipsoid_invcov = a
        self.ellipsoid_cov = cov

        l, v = np.linalg.eigh(a)
        self.ellipsoid_axlens = 1. / np.sqrt(l)
        self.ellipsoid_axes = np.dot(v, np.diag(self.ellipsoid_axlens))

    def update_center(self, ctr):
        """Update ellipsoid center, considering fixed dimensions.

        Parameters
        ----------
        ctr: vector
            new center

        """
        if self.variable_dims is Ellipsis:
            self.ellipsoid_center = ctr
        else:
            self.ellipsoid_center = ctr[self.variable_dims]

    def inside(self, u):
        """Check if inside wrapping ellipsoid.

        Parameters
        ----------
        u: array of vectors
            Points to check

        Returns
        ---------
        is_inside: array of bools
            True if inside wrapping ellipsoid, for each point in `pts`.

        """
        # check the variable subspace with the ellipsoid
        inside_variable = _inside_ellipsoid(u[:,self.variable_dims], self.ellipsoid_center, self.ellipsoid_invcov, self.enlarge)
        if self.variable_dims is Ellipsis:
            return inside_variable
        else:
            # the remaining dims must be exactly equal
            inside_fixed = np.all(self.u[0, ~self.variable_dims] == u[:,~self.variable_dims], axis=1)
            return np.logical_and(inside_fixed, inside_variable)



# cython: language_level=3,annotate=True,profile=True,fast_fail=True,warning_errors=True
"""Efficient Helper functions for stepsamplers
"""

import numpy as np

from numpy import nan as np_nan
import cython
from cython.parallel import prange

@cython.boundscheck(False)
@cython.wraparound(False)
def _within_unit_cube(
    u, 
    acceptable, 
):
    popsize = u.shape[0]
    ndim = u.shape[1]
    # i, j

    for i in range(popsize):
        for j in range(ndim):
            if not 0.0 < u[i,j] < 1.0:
                acceptable[i] = 0
                break


def within_unit_cube(u):
    """whether all fields are between 0 and 1, for each row

    Parameters
    ----------
    u: np.array((npoints, ndim), dtype=float):
        points

    Returns
    ---------
    within: np.array(npoints, dtype=bool):
        for each point, whether it is within the unit cube
    """
    acceptable = np.ones(u.shape[0], dtype=bool)
    _within_unit_cube(u, acceptable)
    return acceptable

@cython.boundscheck(False)
@cython.wraparound(False)
def _evolve_prepare(
     searching_left, 
     searching_right,
     search_right,
     bisecting
):
    # define three mutually exclusive states: 
    # stepping out to the left, to the right, bisecting on the slice
    n = searching_left.shape[0]
    # i
    for i in range(n):
        search_right[i] = not searching_left[i] and searching_right[i]
        bisecting[i] = not (searching_left[i] or searching_right[i])

def evolve_prepare(searching_left, searching_right):
    """Get auxiliary slice sampler state selectors.

    Vectorized computation for multiple (`nwalkers`) walkers.

    Parameters
    ----------
    searching_left: np.array(nwalkers, dtype=bool)
        whether stepping out in the negative direction
    searching_right: np.array(nwalkers, dtype=bool)
        whether stepping out in the positive direction

    Returns
    -------
    search_right: np.array(nwalkers, dtype=bool):
        if searching right and not left
    bisecting: np.array(nwalkers, dtype=bool):
        if not searching right nor left any more
    """
    search_right = np.empty_like(searching_left)
    bisecting = np.empty_like(searching_left)
    _evolve_prepare(searching_left, searching_right, search_right, bisecting)
    return search_right, bisecting

@cython.boundscheck(False)
@cython.wraparound(False)
def evolve_update(
     acceptable, 
     Lnew, 
    Lmin, 
     search_right, 
     bisecting, 
    currentt,
    current_left,
    current_right,
    searching_left,
    searching_right,
    success
):
    """Update the state of each walker.

    This uses the robust logic of slice sampling, 
    with stepping out by doubling.

    Parameters
    ----------
    acceptable: np.array(nwalkers, dtype=bool)
        whether a likelihood evaluation was made. If false, rejected because out of contour.
    Lnew: np.array(acceptable.sum(), dtype=bool)
        likelihood value of proposed point
    Lmin: float
        current log-likelihood threshold
    search_right: np.array(nwalkers, dtype=bool)
        whether stepping out in the positive direction
    bisecting: np.array(nwalkers, dtype=bool)
        whether bisecting. If neither search_right nor bisecting, then 
    currentt: np.array(nwalkers)
        proposed coordinate on the slice
    current_left: np.array(nwalkers)
        current slice negative end
    current_right: np.array(nwalkers)
        current slice positive end
    searching_left: np.array(nwalkers, dtype=bool)
        whether stepping out in the negative direction
    searching_right: np.array(nwalkers, dtype=bool)
        whether stepping out in the positive direction
    success: np.array(nwalkers, dtype=bool)
        whether the walker accepts the point.

    Writes to `currentt`, `current_left`, `current_right`, 
    `searching_left`, `searching_right`, `success`.
    """
    popsize = acceptable.shape[0]
    j = 0
    # i
    my_nan = np_nan
    
    for i in range(popsize):
        if acceptable[i]:
            if Lnew[j] > Lmin:
                success[i] = 1
            j += 1

    for i in prange(popsize, nogil=True):
        # handle cases based on the result:
        # 1) step out further, if still accepting
        if success[i] != 0:
            if searching_left[i]:
                current_left[i] *= 2
            elif search_right[i]:
                current_right[i] *= 2
        # 2) done stepping out, if rejected
        else:
            if searching_left[i]:
                searching_left[i] = 0
            elif search_right[i]:
                searching_right[i] = 0
        # bisecting, rejected or not acceptable
        if bisecting[i]:
            if currentt[i] < 0:
                # bisect shrink left:
                current_left[i] = currentt[i]
            else:
                current_right[i] = currentt[i]
            # bisect accepted: start new slice and new generation there
            if success[i] != 0:
                currentt[i] = my_nan
        else:
            success[i] = 0

# precompute to avoid slow allocations.
pnew_empty = np.empty((0,1))
Lnew_empty = np.empty(0)

def evolve(
    transform, loglike, Lmin, 
    currentu, currentL, currentt, currentv,
    current_left, current_right, searching_left, searching_right,
    log=False
):
    """Evolve each slice sampling walker.

    Parameters
    ----------
    transform: function
        prior transform function
    loglike: function
        loglikelihood function
    Lmin: float
        current log-likelihood threshold
    search_right: np.array(nwalkers, dtype=bool)
        whether stepping out in the positive direction
    bisecting: np.array(nwalkers, dtype=bool)
        whether bisecting. If neither search_right nor bisecting, then 
    currentu: np.array((nwalkers, ndim))
        slice starting point (where currentt=0)
    currentL: np.array(nwalkers)
        current loglikelihood
    currentt: np.array(nwalkers)
        proposed coordinate on the slice
    currentv: np.array((nwalkers, ndim))
        slice direction vector
    current_left: np.array(nwalkers)
        current slice negative end
    current_right: np.array(nwalkers)
        current slice positive end
    searching_left: np.array(nwalkers, dtype=bool)
        whether stepping out in the negative direction
    searching_right: np.array(nwalkers, dtype=bool)
        whether stepping out in the positive direction

    Returns
    -------
    currentt: np.array(nwalkers)
        as above
    currentv: np.array((nwalkers, ndim))
        as above
    current_left: np.array(nwalkers)
        as above
    current_right: np.array(nwalkers)
        as above
    searching_left: np.array(nwalkers, dtype=bool)
        as above
    searching_right: np.array(nwalkers, dtype=bool)
        as above
    success: np.array(nwalkers, dtype=bool)
        whether the walker accepts the point.
    unew: np.array((success.sum(), ndim))
        coordinates of accepted points
    pnew: np.array((success.sum(), nparams))
        transformed coordinates of accepted points
    Lnew: np.array(success.sum())
        log-likelihoods of accepted points
    nc: int
        number of points for which the log-likelihood function was called.

    This function writes in-place to 
    `currentt`, `currentv`, `current_left`, `current_right`, `searching_left`, 
    `searching_right` and `currentu`, but also returns these.
    """
    search_right, bisecting = evolve_prepare(searching_left, searching_right)

    unew = currentu
    unew[searching_left,:] = currentu[searching_left,:] + currentv[searching_left,:] * current_left[searching_left].reshape((-1,1))
    unew[search_right,:] = currentu[search_right,:] + currentv[search_right,:] * current_right[search_right].reshape((-1,1))
    currentt[bisecting] = np.random.uniform(current_left[bisecting], current_right[bisecting])
    unew[bisecting,:] = currentu[bisecting,:] + currentv[bisecting,:] * currentt[bisecting].reshape((-1,1))

    acceptable = within_unit_cube(unew)

    nc = 0
    if acceptable.any():
        pnew = transform(unew[acceptable,:])
        Lnew = loglike(pnew)
        nc += len(pnew)
    else:
        pnew = pnew_empty
        Lnew = Lnew_empty

    success = np.zeros_like(searching_left)
    evolve_update(
        acceptable, Lnew, Lmin, search_right, bisecting, currentt,
        current_left, current_right, searching_left, searching_right,
        success
    )

    return (
        (
        currentt, currentv,
        current_left, current_right, searching_left, searching_right), 
        (success, unew[success,:], pnew[success[acceptable],:], Lnew[success[acceptable]]), 
        nc
    )

def step_back(Lmin, allL, generation, currentt, log=False):
    """Revert walkers which have wandered astray.

    Revert until all previous steps have likelihoods allL above Lmin.
    Updates currentt, generation and allL, in-place.

    Parameters
    ----------
    Lmin: float
        current loglikelihood threshold
    allL: np.array((nwalkers, ngenerations))
        loglikelihoods of the chain. NaN where not evaluated yet.
    generation: np.array(nwalkers, dtype=int)
        how many iterations each walker has completed.
    currentt: np.array(nwalkers)
        current slice coordinate

    """
    # step back where step was excluded by Lmin increase
    # delete from the back until all are good:
    max_width = generation.max() + 1
    below_threshold = allL[:,:max_width] < Lmin
    problematic_parent = np.any(below_threshold, axis=1)
    if not problematic_parent.any():
        return
    parent_i, = np.where(problematic_parent)
    below_threshold_parent = below_threshold[parent_i,:]
    # first, all of them (because we already identified them)
    problematic = np.ones(len(parent_i), dtype=bool)
    step = 0

    while True:
        step += 1
        ii, = np.where(problematic)
        i = parent_i[problematic]
        g = generation[i]
        generation[i] -= 1
        currentt[i] = np_nan
        allL[i,g] = np_nan
        below_threshold_parent[problematic, g] = False
        if log:
            print("resetting %d%%" % (problematic.meancount_good_generations() * 100), 'by', step, 'steps', 'to', g)

        problematic = np.any(below_threshold_parent, axis=1)
        if not problematic.any():
            break


def _fill_directions(
     v,
     indices,
     scale
):
    nsamples = v.shape[0]
    # i
    for i in range(nsamples):
        v[i, indices[i]] = scale

def generate_cube_oriented_direction(ui, region, scale=1):
    """Draw a unit direction vector in direction of a random unit cube axes.

    Parameters
    ----------
    ui: np.array((npoints, ndim), dtype=float)
        starting points (not used)
    region:
        not used
    scale: float
        length of returned vector

    Returns
    ---------
    v: np.array((npoints, ndim), dtype=float)
        Random axis vectors of length `scale`, one for each starting point.
    """
    nsamples, ndim = ui.shape
    v = np.zeros((nsamples, ndim))
    # choose axis
    j = np.random.randint(ndim, size=nsamples)
    _fill_directions(v, j, scale)
    return v

def generate_random_direction(ui, region, scale=1):
    """Draw uniform direction vector in unit cube space of length `scale`.

    Parameters
    -----------
    region: MLFriends object
        current region (not used)
    scale: float
        length of direction vector
    
    Returns
    --------
    v: array
        new direction vector
    """
    del region
    nsamples, ndim = ui.shape
    v = np.random.normal(size=(nsamples, ndim))
    v *= scale / np.linalg.norm(v, axis=1).reshape((nsamples, 1))
    return v


def generate_region_oriented_direction(ui, region, scale=1):
    """Draw a random direction vector in direction of one of the `region` axes.

    If given, the vector length is `scale`.
    If not, the vector length in transformed space is `tscale`.

    Parameters
    -----------
    region: MLFriends object
        current region
    scale: float
        length of direction vector in t-space

    Returns
    --------
    v: array
        new direction vector (in u-space)
    """
    nsamples, ndim = ui.shape
    # choose axis in transformed space:
    j = np.random.randint(ndim, size=nsamples)
    v = region.transformLayer.axes[j] * scale
    return v


def generate_region_random_direction(ui, region, scale=1):
    """Draw a direction vector in a random direction of the region.

    The vector length is `scale` (in unit cube space).

    Parameters
    -----------
    region: MLFriends object
        current region
    scale: float:
        length of direction vector (in t-space)
    
    Returns
    --------
    v: array
        new direction vector
    """
    nsamples, ndim = ui.shape
    # choose axis in transformed space:
    v1 = np.random.normal(size=(nsamples, ndim))
    v1 *= scale / np.linalg.norm(v1, axis=1).reshape((nsamples, 1))
    v = np.einsum('ij,kj->ki', region.transformLayer.axes, v1)
    return v



import numpy as np
import shutil
import tempfile
import pytest
from numpy.testing import assert_allclose

def test_run():
    from ultranest import NestedSampler

    def loglike(y):
        z = np.log10(y)
        a = np.array([-0.5 * sum([((xi - 0.83456 + i*0.1)/0.5)**2 for i, xi in enumerate(x)]) for x in z])
        b = np.array([-0.5 * sum([((xi - 0.43456 - i*0.1)/0.5)**2 for i, xi in enumerate(x)]) for x in z])
        loglike.ncalls += len(a)
        return np.logaddexp(a, b)
    loglike.ncalls = 0

    def transform(x):
        return 10**(10. * x - 5.)

    paramnames = ['Hinz', 'Kunz']

    sampler = NestedSampler(paramnames, loglike, transform=transform, num_live_points=400, vectorized=True)
    r = sampler.run(log_interval=50)

    ncalls = loglike.ncalls
    if sampler.mpi_size > 1:
        ncalls = sampler.comm.gather(ncalls, root=0)
        if sampler.mpi_rank == 0:
            print("ncalls on the different MPI ranks:", ncalls)
        ncalls = sum(sampler.comm.bcast(ncalls, root=0))
    assert abs(r['ncall'] - ncalls) <= 2 * sampler.mpi_size, (r['ncall'], ncalls)
    open('nestedsampling_results.txt', 'a').write("%.3f\n" % r['logz'])
    sampler.plot()


def test_dlogz_reactive_run_SLOW():
    from ultranest import ReactiveNestedSampler
    import ultranest.mlfriends

    def loglike(y):
        return -0.5 * np.sum(((y - 0.5)/0.001)**2, axis=1)

    paramnames = ['Hinz', 'Kunz']

    sampler = ReactiveNestedSampler(paramnames, loglike, vectorized=True)
    print("running for ess")
    firstresults = sampler.run(min_num_live_points=50, cluster_num_live_points=0, max_num_improvement_loops=3, min_ess=10000, viz_callback=None)
    print()
    print({k:v for k, v in firstresults.items() if 'logzerr' in k})
    print()
    assert firstresults['logzerr'] > 0.1 * 2
    print("running again for logz")
    for niter, results in enumerate(sampler.run_iter(min_num_live_points=1, cluster_num_live_points=0, max_num_improvement_loops=10, dlogz=0.1, viz_callback=None, region_class=ultranest.mlfriends.RobustEllipsoidRegion)):
        print("logzerr in iteration %d" % niter, results['logzerr'])
    print()
    print({k:v for k, v in results.items() if 'logzerr' in k})
    assert results['logzerr'] < 0.1 * 2

def test_reactive_run():
    from ultranest import ReactiveNestedSampler
    np.random.seed(1)
    evals = set()

    def loglike(z):
        #print(loglike.ncalls, z[0,0])
        [evals.add(str(x[0])) for x in z]
        a = np.array([-0.5 * sum([((xi - 0.83456 + i*0.1)/0.5)**2 for i, xi in enumerate(x)]) for x in z])
        b = np.array([-0.5 * sum([((xi - 0.43456 - i*0.1)/0.5)**2 for i, xi in enumerate(x)]) for x in z])
        loglike.ncalls += len(a)
        return np.logaddexp(a, b)
    loglike.ncalls = 0

    def transform(x):
        return 10. * x - 5.

    paramnames = ['Hinz', 'Kunz']

    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform,
        draw_multiple=False, vectorized=True)
    r = sampler.run(log_interval=50, min_num_live_points=400)

    # test that the number of likelihood calls is correct

    ncalls = loglike.ncalls
    nunique = len(evals)
    if sampler.mpi_size > 1:
        ncalls = sampler.comm.gather(ncalls, root=0)
        if sampler.mpi_rank == 0:
            print("ncalls on the different MPI ranks:", ncalls)
        ncalls = sum(sampler.comm.bcast(ncalls, root=0))

        allevals = sampler.comm.gather(evals, root=0)
        if sampler.mpi_rank == 0:
            print("evals on the different MPI ranks:", [len(e) for e in allevals])
            allevals = len(set.union(*allevals))
        else:
            allevals = None
        nunique = sampler.comm.bcast(allevals, root=0)

    if sampler.mpi_rank == 0:
        print('ncalls:', ncalls, 'nunique:', nunique)

    assert abs(r['ncall'] - ncalls) <= 2 * sampler.mpi_size, (r['ncall'], ncalls)
    assert ncalls == nunique, (ncalls, nunique)
    if sampler.mpi_rank == 0:
        open('nestedsampling_reactive_results.txt', 'a').write("%.3f\n" % r['logz'])

    print(r)
    assert r['niter'] > 100
    assert -10 < r['logz'] < 10
    assert 0.01 < r['logzerr'] < 0.5
    assert 1 < r['ess'] < 10000


    sampler.plot()


def test_reactive_run_extraparams():
    from ultranest import ReactiveNestedSampler
    np.random.seed(1)

    def loglike(z):
        return -0.5 * z[-1].sum()
    loglike.ncalls = 0

    def transform(x):
        z = 10. * x - 5.
        return np.append(z, np.abs(z).sum())

    paramnames = ['Hinz', 'Kunz']

    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform,
        derived_param_names=['ctr_distance'])
    r = sampler.run()
    assert r['samples'].shape[1] == 3
    sampler.plot()

def test_return_summary():
    from ultranest import ReactiveNestedSampler
    sigma = np.array([0.1, 0.01])
    centers = np.array([0.5, 0.75])
    paramnames = ['a', 'b']
    ndim = len(paramnames)

    def loglike(theta):
        like = -0.5 * (((theta - centers)/sigma)**2) - 0.5 * np.log(2 * np.pi * sigma**2) * ndim
        return like.sum()

    def transform(x):
        return x

    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform)
    r = sampler.run()

    print(r)
    assert r['paramnames'] == paramnames
    assert r['niter'] > 100
    assert -10 < r['logz'] < 10
    assert 0.01 < r['logzerr'] < 0.5
    assert 1 < r['ess'] < 10000
    assert 0.4 < r['posterior']['mean'][0] < 0.6
    assert 0.74 < r['posterior']['mean'][1] < 0.76
    assert 0.4 < r['posterior']['median'][0] < 0.6
    assert 0.74 < r['posterior']['median'][1] < 0.76
    assert 0.05 < r['posterior']['stdev'][0] < 0.2
    assert 0.005 < r['posterior']['stdev'][1] < 0.02

    assert 0.35 < r['posterior']['errlo'][0] < 0.45
    assert 0.72 < r['posterior']['errlo'][1] < 0.75
    assert 0.55 < r['posterior']['errup'][0] < 0.65
    assert 0.75 < r['posterior']['errup'][1] < 0.78

    N, ndim2 = r['samples'].shape
    assert ndim2 == ndim
    assert N > 10
    N, ndim2 = r['weighted_samples']['points'].shape
    assert ndim2 == ndim
    assert N > 10

    assert r['weighted_samples']['logw'].shape == (N,)
    assert r['weighted_samples']['weights'].shape == (N,)
    assert r['weighted_samples']['bootstrapped_weights'].shape[0] == N
    assert r['weighted_samples']['logl'].shape == (N,)

@pytest.mark.parametrize("dlogz", [2.0, 0.5, 0.1])
def test_run_resume(dlogz):
    from ultranest import ReactiveNestedSampler
    sigma = 0.01
    ndim = 1

    def loglike(theta):
        like = -0.5 * (((theta - 0.5)/sigma)**2).sum(axis=1) - 0.5 * np.log(2 * np.pi * sigma**2) * ndim
        return like

    def transform(x):
        return x

    paramnames = ['a']
    def myadd(row):
        assert False, (row, 'should not need to add more points in resume')

    last_results = None
    #for dlogz in 0.5, 0.1, 0.01:
    np.random.seed(int(dlogz*100))
    folder = tempfile.mkdtemp()
    try:
        for i in range(2):
            sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform,
                log_dir=folder, resume=True, vectorized=True)
            r = sampler.run(log_interval=50, dlogz=dlogz, min_num_live_points=400)
            sampler.print_results()
            sampler.pointstore.close()
            if i == 1:
                sampler.pointstore.add = myadd
            del r['weighted_samples']
            del r['samples']
            if last_results is not None:
                print("ran with dlogz:", dlogz)
                print("first run gave:", last_results)
                print("second run gave:", r)
                assert last_results['logzerr'] < 1.0
                assert r['logzerr'] < 1.0
                assert np.isclose(last_results['logz'], r['logz'], atol=0.5)
            last_results = r
    finally:
        shutil.rmtree(folder, ignore_errors=True)

@pytest.mark.parametrize("storage_backend", ['hdf5', 'tsv', 'csv'])
def test_reactive_run_resume_eggbox(storage_backend):
    from ultranest import ReactiveNestedSampler
    from ultranest import read_file

    def loglike(z):
        chi = (np.cos(z / 2.)).prod(axis=1)
        loglike.ncalls += len(z)
        return (2. + chi)**5
    loglike.ncalls = 0

    def transform(x):
        return x * 10 * np.pi

    paramnames = ['a', 'b']
    ndim = len(paramnames)

    #last_results = None
    folder = tempfile.mkdtemp()
    np.random.seed(1)
    try:
        for i in range(2):
            print()
            print("====== Running Eggbox problem [%d] =====" % (i+1))
            print()
            sampler = ReactiveNestedSampler(paramnames,
                loglike, transform=transform,
                log_dir=folder, resume=True, vectorized=True, draw_multiple=False,
                storage_backend=storage_backend)
            initial_ncalls = int(sampler.ncall)
            num_live_points = 100
            loglike.ncalls = 0
            r = sampler.run(max_iters=200 + i*200,
                max_num_improvement_loops=0,
                min_num_live_points=num_live_points,
                cluster_num_live_points=0)
            sampler.print_results()
            if storage_backend == 'hdf5':
                print("pointstore:", sampler.pointstore.fileobj['points'].shape)
            sampler.pointstore.close()
            print(loglike.ncalls, r['ncall'], initial_ncalls)

            ncalls = loglike.ncalls
            if sampler.mpi_size > 1:
                ncalls = sampler.comm.gather(ncalls, root=0)
                if sampler.mpi_rank == 0:
                    print("ncalls on the different MPI ranks:", ncalls)
                ncalls = sum(sampler.comm.bcast(ncalls, root=0))
            ncalls = ncalls + initial_ncalls
            assert abs(r['ncall'] - ncalls) <= 2 * sampler.mpi_size, (i, r['ncall'], ncalls, r['ncall'] - ncalls)
            assert paramnames == r['paramnames'], 'paramnames should be in results'

        # the results are not exactly the same, because the sampling adds
        #ncalls = loglike.ncalls
        #sampler = ReactiveNestedSampler(paramnames,
        #    loglike, transform=transform,
        #    log_dir=folder, resume=True, vectorized=True, num_test_samples=0)
        #print("pointstore:", sampler.pointstore.fileobj['points'].shape)
        #assert ncalls == loglike.ncalls, (ncalls, loglike.ncalls)
        if storage_backend == 'hdf5':
            sequence, results = read_file(folder, ndim, random=False, num_bootstraps=0)

            print("sampler results: ********************")
            print({k:v for k, v in r.items() if np.asarray(v).size < 20 and k != 'weighted_samples'})
            print("reader results: ********************")
            print({k:v for k, v in results.items() if np.asarray(v).size < 20 and k != 'weighted_samples'})
            for k, v in results.items():
                if k == 'posterior' or k == 'samples':
                    pass
                elif k == 'weighted_samples' or k == 'maximum_likelihood':
                    for k2, v2 in results[k].items():
                        if k2 == 'bootstrapped_weights': continue
                        print("  ", k, "::", k2, np.shape(v2))
                        assert_allclose(r[k][k2], v2)
                elif k.startswith('logzerr') or '_bs' in k or 'Herr' in k:
                    print("   skipping", k, np.shape(v))
                    #assert_allclose(r[k], v, atol=0.5)
                elif k == 'insertion_order_MWW_test':
                    print('insertion_order_MWW_test:', r[k], v)
                    assert r[k] == v, (r[k], v)
                else:
                    print("  ", k, np.shape(v))
                    assert_allclose(r[k], v)

            logw = r['weighted_samples']['logw']
            v = r['weighted_samples']['points']
            L = r['weighted_samples']['logl']

            assert sequence['logz'][-1] - r['logz'] < 0.5, (results['logz'][-1], r['logz'])
            assert sequence['logzerr'][-1] <= r['logzerr_single'], (results['logzerr'][-1], r['logzerr'])
            #assert_allclose(sequence['logz_final'], r['logz_single'], atol=0.3)
            #assert_allclose(sequence['logzerr_final'], r['logzerr_single'], atol=0.1)
            assert r['niter'] <= sequence['niter'] <= r['niter'], (sequence['niter'], r['niter'])
            assert results['niter'] == len(sequence['logz']) == len(sequence['logzerr']) == len(sequence['logvol']) == len(sequence['logwt'])
            assert results['niter'] == len(results['samples'])
            data = np.loadtxt(folder + '/chains/weighted_post.txt', skiprows=1)
            assert_allclose(data[:,0], results['weighted_samples']['weights'])
            assert_allclose(data[:,1], results['weighted_samples']['logl'])
            assert_allclose(v, results['weighted_samples']['points'])
            assert_allclose(logw, results['weighted_samples']['logw'])
            assert_allclose(L, results['weighted_samples']['logl'])

            assert_allclose(L, sequence['logl'])
            #assert_allclose(logw + L, sequence['logwt'])
            assert sequence['logvol'].shape == logw.shape == (len(L),), (sequence['logvol'].shape, logw.shape)
            assert sequence['logwt'].shape == logw.shape == (len(L),), (sequence['logwt'].shape, logw.shape)
            #assert_allclose(logw, sequence['logvols'])
            #assert results['samples_untransformed'].shape == v.shape == (len(L), ndim), (results['samples_untransformed'].shape, v.shape)

    finally:
        shutil.rmtree(folder, ignore_errors=True)

def test_reactive_run_warmstart_gauss():
    from ultranest import ReactiveNestedSampler
    from ultranest import read_file
    center = 0

    def loglike(z):
        chi2 = (((z - center)/0.001)**2).sum(axis=1)
        loglike.ncalls += len(z)
        return -0.5 * chi2
    loglike.ncalls = 0

    def transform(x):
        return x * 20000 - 10000

    paramnames = ['a']
    ndim = len(paramnames)

    folder = tempfile.mkdtemp()
    np.random.seed(1)
    first_ncalls = None
    resume_ncalls = None
    try:
        for i, resume in enumerate(['overwrite', 'resume', 'resume-similar']):
            print()
            print("====== Running Gauss problem [%d] =====" % (i+1))
            print()
            center = (i+1) * 1e-4
            try:
                sampler = ReactiveNestedSampler(paramnames,
                    loglike, transform=transform,
                    log_dir=folder, resume=resume, vectorized=True, draw_multiple=False,
                    warmstart_max_tau=0.5)
            except Exception as e:
                # we expect an error for resuming with a changed likelihood
                if resume != 'resume':
                    raise e
                else:
                    assert 'loglikelihood function changed' in str(e), e
                    print("Exception as expected:", e)
                    continue
            initial_ncalls = int(sampler.ncall)
            if i == 0:
                assert initial_ncalls == 0
            num_live_points = 100
            loglike.ncalls = 0
            r = sampler.run(
                max_num_improvement_loops=0,
                min_num_live_points=num_live_points,
                cluster_num_live_points=0, viz_callback=None, frac_remain=
                0.5)
            sampler.print_results()
            print("pointstore:", sampler.pointstore.fileobj['points'].shape)
            sampler.pointstore.close()
            print(loglike.ncalls, r['ncall'], initial_ncalls)

            ncalls = loglike.ncalls
            if sampler.mpi_size > 1:
                ncalls = sampler.comm.gather(ncalls, root=0)
                if sampler.mpi_rank == 0:
                    print("ncalls on the different MPI ranks:", ncalls)
                ncalls = sum(sampler.comm.bcast(ncalls, root=0))
            ncalls = ncalls + initial_ncalls
            if i == 0:
                first_ncalls = ncalls
            if i == 2:
                resume_ncalls = loglike.ncalls
            assert abs(r['ncall'] - ncalls) <= 2 * sampler.mpi_size, (i, r['ncall'], ncalls, r['ncall'] - ncalls)
            assert paramnames == r['paramnames'], 'paramnames should be in results'

    finally:
        shutil.rmtree(folder, ignore_errors=True)
    
    # make sure warm start is much faster
    assert resume_ncalls < first_ncalls - 800, (resume_ncalls, first_ncalls)

def test_run_compat():
    from ultranest.solvecompat import pymultinest_solve_compat as solve

    ndim = 2
    sigma = 0.01
    centers = 0.5
    paramnames = ['a', 'b']

    def loglike(theta):
        like = -0.5 * (((theta - centers)/sigma)**2).sum() - 0.5 * np.log(2 * np.pi * sigma**2) * ndim
        return like

    def transform(x):
        params = x.copy()
        params[0] = 10 * x[0] - 5.
        params[1] = 10**(x[1] - 1)
        return params

    result = solve(LogLikelihood=loglike, Prior=transform,
        n_dims=ndim, outputfiles_basename=None,
        verbose=True, resume='resume', importance_nested_sampling=False)

    print()
    print('evidence: %(logZ).1f +- %(logZerr).1f' % result)
    print()
    print('parameter values:')
    for name, col in zip(paramnames, result['samples'].transpose()):
        print('%15s : %.3f +- %.3f' % (name, col.mean(), col.std()))

if __name__ == '__main__':
    #test_run_compat()
    #test_run_resume(dlogz=0.5)
    #test_reactive_run_resume(dlogz=0.5, min_ess=1000)
    #test_reactive_run()
    #test_run()
    #test_reactive_run_warmstart_gauss()
    #test_reactive_run_extraparams()
    test_dlogz_reactive_run()
